#-------------------------------------------------------------------------------
# elftools: dwarf/enums.py
#
# Mappings of enum names to values
#
# Eli Bendersky (eliben@gmail.com)
# This code is in the public domain
#-------------------------------------------------------------------------------
from ..construct import Pass


ENUM_DW_TAG = dict(
    DW_TAG_null                        = 0x00,
    DW_TAG_array_type                  = 0x01,
    DW_TAG_class_type                  = 0x02,
    DW_TAG_entry_point                 = 0x03,
    DW_TAG_enumeration_type            = 0x04,
    DW_TAG_formal_parameter            = 0x05,
    DW_TAG_global_subroutine           = 0x06,
    DW_TAG_global_variable             = 0x07,
    DW_TAG_imported_declaration        = 0x08,
    DW_TAG_label                       = 0x0a,
    DW_TAG_lexical_block               = 0x0b,
    DW_TAG_local_variable              = 0x0c,
    DW_TAG_member                      = 0x0d,
    DW_TAG_pointer_type                = 0x0f,
    DW_TAG_reference_type              = 0x10,
    DW_TAG_compile_unit                = 0x11,
    DW_TAG_string_type                 = 0x12,
    DW_TAG_structure_type              = 0x13,
    DW_TAG_subroutine                  = 0x14,
    DW_TAG_subroutine_type             = 0x15,
    DW_TAG_typedef                     = 0x16,
    DW_TAG_union_type                  = 0x17,
    DW_TAG_unspecified_parameters      = 0x18,
    DW_TAG_variant                     = 0x19,
    DW_TAG_common_block                = 0x1a,
    DW_TAG_common_inclusion            = 0x1b,
    DW_TAG_inheritance                 = 0x1c,
    DW_TAG_inlined_subroutine          = 0x1d,
    DW_TAG_module                      = 0x1e,
    DW_TAG_ptr_to_member_type          = 0x1f,
    DW_TAG_set_type                    = 0x20,
    DW_TAG_subrange_type               = 0x21,
    DW_TAG_with_stmt                   = 0x22,
    DW_TAG_access_declaration          = 0x23,
    DW_TAG_base_type                   = 0x24,
    DW_TAG_catch_block                 = 0x25,
    DW_TAG_const_type                  = 0x26,
    DW_TAG_constant                    = 0x27,
    DW_TAG_enumerator                  = 0x28,
    DW_TAG_file_type                   = 0x29,
    DW_TAG_friend                      = 0x2a,
    DW_TAG_namelist                    = 0x2b,
    DW_TAG_namelist_item               = 0x2c,
    DW_TAG_namelist_items              = 0x2c,
    DW_TAG_packed_type                 = 0x2d,
    DW_TAG_subprogram                  = 0x2e,

    # The DWARF standard defines these as _parameter, not _param, but we
    # maintain compatibility with readelf.
    DW_TAG_template_type_param         = 0x2f,
    DW_TAG_template_value_param        = 0x30,

    DW_TAG_thrown_type                 = 0x31,
    DW_TAG_try_block                   = 0x32,
    DW_TAG_variant_part                = 0x33,
    DW_TAG_variable                    = 0x34,
    DW_TAG_volatile_type               = 0x35,
    DW_TAG_dwarf_procedure             = 0x36,
    DW_TAG_restrict_type               = 0x37,
    DW_TAG_interface_type              = 0x38,
    DW_TAG_namespace                   = 0x39,
    DW_TAG_imported_module             = 0x3a,
    DW_TAG_unspecified_type            = 0x3b,
    DW_TAG_partial_unit                = 0x3c,
    DW_TAG_imported_unit               = 0x3d,
    DW_TAG_mutable_type                = 0x3e,
    DW_TAG_condition                   = 0x3f,
    DW_TAG_shared_type                 = 0x40,
    DW_TAG_type_unit                   = 0x41,
    DW_TAG_rvalue_reference_type       = 0x42,
    DW_TAG_template_alias              = 0x43,
    DW_TAG_coarray_type                = 0x44,
    DW_TAG_generic_subrange            = 0x45,
    DW_TAG_dynamic_type                = 0x46,    
    DW_TAG_atomic_type                 = 0x47,
    DW_TAG_call_site                   = 0x48,
    DW_TAG_call_site_parameter         = 0x49,
    DW_TAG_skeleton_unit               = 0x4a,
    DW_TAG_immutable_type              = 0x4b,

    # Tags between 0x4080 and 0xffff are user-defined.
    # different implementations may overlap?

    DW_TAG_lo_user                     = 0x4080,
    DW_TAG_GNU_template_template_param = 0x4106,
    DW_TAG_GNU_template_parameter_pack = 0x4107,
    DW_TAG_GNU_formal_parameter_pack   = 0x4108,
    DW_TAG_GNU_call_site               = 0x4109,
    DW_TAG_GNU_call_site_parameter     = 0x410a,

    DW_TAG_APPLE_property              = 0x4200,

    DW_TAG_hi_user                     = 0xffff,

    _default_                          = Pass,
)


ENUM_DW_CHILDREN = dict(
    DW_CHILDREN_no  = 0x00,
    DW_CHILDREN_yes = 0x01,
)


ENUM_DW_AT = dict(
    DW_AT_null                      = 0x00,
    DW_AT_sibling                   = 0x01,
    DW_AT_location                  = 0x02,
    DW_AT_name                      = 0x03,
    DW_AT_fund_type                 = 0x05,
    DW_AT_mod_fund_type             = 0x06,
    DW_AT_user_def_type             = 0x07,
    DW_AT_mod_u_d_type              = 0x08,
    DW_AT_ordering                  = 0x09,
    DW_AT_subscr_data               = 0x0a,
    DW_AT_byte_size                 = 0x0b,
    DW_AT_bit_offset                = 0x0c,
    DW_AT_bit_size                  = 0x0d,
    DW_AT_element_list              = 0x0f,
    DW_AT_stmt_list                 = 0x10,
    DW_AT_low_pc                    = 0x11,
    DW_AT_high_pc                   = 0x12,
    DW_AT_language                  = 0x13,
    DW_AT_member                    = 0x14,
    DW_AT_discr                     = 0x15,
    DW_AT_discr_value               = 0x16,
    DW_AT_visibility                = 0x17,
    DW_AT_import                    = 0x18,
    DW_AT_string_length             = 0x19,
    DW_AT_common_reference          = 0x1a,
    DW_AT_comp_dir                  = 0x1b,
    DW_AT_const_value               = 0x1c,
    DW_AT_containing_type           = 0x1d,
    DW_AT_default_value             = 0x1e,
    DW_AT_friends                   = 0x1f,
    DW_AT_inline                    = 0x20,
    DW_AT_is_optional               = 0x21,
    DW_AT_lower_bound               = 0x22,
    DW_AT_program                   = 0x23,
    DW_AT_private                   = 0x24,
    DW_AT_producer                  = 0x25,
    DW_AT_protected                 = 0x26,
    DW_AT_prototyped                = 0x27,
    DW_AT_public                    = 0x28,
    DW_AT_pure_virtual              = 0x29,
    DW_AT_return_addr               = 0x2a,
    # In DWARFv1, DW_AT_specification was at 0x2b, moved to 0x47 in v2
    DW_AT_start_scope               = 0x2c,
    DW_AT_bit_stride                = 0x2e,
    DW_AT_stride_size               = 0x2e,
    DW_AT_upper_bound               = 0x2f,
    DW_AT_virtual                   = 0x30,
    DW_AT_abstract_origin           = 0x31,
    DW_AT_accessibility             = 0x32,
    DW_AT_address_class             = 0x33,
    DW_AT_artificial                = 0x34,
    DW_AT_base_types                = 0x35,
    DW_AT_calling_convention        = 0x36,
    DW_AT_count                     = 0x37,
    DW_AT_data_member_location      = 0x38,
    DW_AT_decl_column               = 0x39,
    DW_AT_decl_file                 = 0x3a,
    DW_AT_decl_line                 = 0x3b,
    DW_AT_declaration               = 0x3c,
    DW_AT_discr_list                = 0x3d,
    DW_AT_encoding                  = 0x3e,
    DW_AT_external                  = 0x3f,
    DW_AT_frame_base                = 0x40,
    DW_AT_friend                    = 0x41,
    DW_AT_identifier_case           = 0x42,
    DW_AT_macro_info                = 0x43,
    DW_AT_namelist_item             = 0x44,
    DW_AT_priority                  = 0x45,
    DW_AT_segment                   = 0x46,
    DW_AT_specification             = 0x47,
    DW_AT_static_link               = 0x48,
    DW_AT_type                      = 0x49,
    DW_AT_use_location              = 0x4a,
    DW_AT_variable_parameter        = 0x4b,
    DW_AT_virtuality                = 0x4c,
    DW_AT_vtable_elem_location      = 0x4d,
    DW_AT_allocated                 = 0x4e,
    DW_AT_associated                = 0x4f,
    DW_AT_data_location             = 0x50,
    DW_AT_byte_stride               = 0x51,
    DW_AT_stride                    = 0x51,
    DW_AT_entry_pc                  = 0x52,
    DW_AT_use_UTF8                  = 0x53,
    DW_AT_extension                 = 0x54,
    DW_AT_ranges                    = 0x55,
    DW_AT_trampoline                = 0x56,
    DW_AT_call_column               = 0x57,
    DW_AT_call_file                 = 0x58,
    DW_AT_call_line                 = 0x59,
    DW_AT_description               = 0x5a,
    DW_AT_binary_scale              = 0x5b,
    DW_AT_decimal_scale             = 0x5c,
    DW_AT_small                     = 0x5d,
    DW_AT_decimal_sign              = 0x5e,
    DW_AT_digit_count               = 0x5f,
    DW_AT_picture_string            = 0x60,
    DW_AT_mutable                   = 0x61,
    DW_AT_threads_scaled            = 0x62,
    DW_AT_explicit                  = 0x63,
    DW_AT_object_pointer            = 0x64,
    DW_AT_endianity                 = 0x65,
    DW_AT_elemental                 = 0x66,
    DW_AT_pure                      = 0x67,
    DW_AT_recursive                 = 0x68,
    DW_AT_signature                 = 0x69,
    DW_AT_main_subprogram           = 0x6a,
    DW_AT_data_bit_offset           = 0x6b,
    DW_AT_const_expr                = 0x6c,
    DW_AT_enum_class                = 0x6d,
    DW_AT_linkage_name              = 0x6e,
    DW_AT_string_length_bit_size    = 0x6f,
    DW_AT_string_length_byte_size   = 0x70,
    DW_AT_rank                      = 0x71,
    DW_AT_str_offsets_base          = 0x72,
    DW_AT_addr_base                 = 0x73,
    DW_AT_rnglists_base             = 0x74,
    DW_AT_dwo_name                  = 0x76,
    DW_AT_reference                 = 0x77,
    DW_AT_rvalue_reference          = 0x78,
    DW_AT_macros                    = 0x79,
    DW_AT_call_all_calls            = 0x7a,
    DW_AT_call_all_source_calls     = 0x7b,
    DW_AT_call_all_tail_calls       = 0x7c,
    DW_AT_call_return_pc            = 0x7d,
    DW_AT_call_value                = 0x7e,
    DW_AT_call_origin               = 0x7f,
    DW_AT_call_parameter            = 0x80,
    DW_AT_call_pc                   = 0x81,
    DW_AT_call_tail_call            = 0x82,
    DW_AT_call_target               = 0x83,
    DW_AT_call_target_clobbered     = 0x84,
    DW_AT_call_data_location        = 0x85,
    DW_AT_call_data_value           = 0x86,
    DW_AT_noreturn                  = 0x87,
    DW_AT_alignment                 = 0x88,
    DW_AT_export_symbols            = 0x89,
    DW_AT_deleted                   = 0x8a,
    DW_AT_defaulted                 = 0x8b,
    DW_AT_loclists_base             = 0x8c,

    DW_AT_MIPS_fde                      = 0x2001,
    DW_AT_MIPS_loop_begin               = 0x2002,
    DW_AT_MIPS_tail_loop_begin          = 0x2003,
    DW_AT_MIPS_epilog_begin             = 0x2004,
    DW_AT_MIPS_loop_unroll_factor       = 0x2005,
    DW_AT_MIPS_software_pipeline_depth  = 0x2006,
    DW_AT_MIPS_linkage_name             = 0x2007,
    DW_AT_MIPS_stride                   = 0x2008,
    DW_AT_MIPS_abstract_name            = 0x2009,
    DW_AT_MIPS_clone_origin             = 0x200a,
    DW_AT_MIPS_has_inlines              = 0x200b,
    DW_AT_MIPS_stride_byte              = 0x200c,
    DW_AT_MIPS_stride_elem              = 0x200d,
    DW_AT_MIPS_ptr_dopetype             = 0x200e,
    DW_AT_MIPS_allocatable_dopetype     = 0x200f,
    DW_AT_MIPS_assumed_shape_dopetype   = 0x2010,
    DW_AT_MIPS_assumed_size             = 0x2011,

    DW_AT_HP_opt_level                  = 0x2014,

    DW_AT_sf_names                      = 0x2101,
    DW_AT_src_info                      = 0x2102,
    DW_AT_mac_info                      = 0x2103,
    DW_AT_src_coords                    = 0x2104,
    DW_AT_body_begin                    = 0x2105,
    DW_AT_body_end                      = 0x2106,
    DW_AT_GNU_vector                    = 0x2107,
    DW_AT_GNU_template_name             = 0x2110,
    DW_AT_GNU_odr_signature             = 0x210f,

    DW_AT_GNU_call_site_value               = 0x2111,
    DW_AT_GNU_call_site_data_value          = 0x2112,
    DW_AT_GNU_call_site_target              = 0x2113,
    DW_AT_GNU_call_site_target_clobbered    = 0x2114,
    DW_AT_GNU_tail_call                     = 0x2115,
    DW_AT_GNU_all_tail_call_sites           = 0x2116,
    DW_AT_GNU_all_call_sites                = 0x2117,
    DW_AT_GNU_all_source_call_sites         = 0x2118,
    DW_AT_GNU_macros                        = 0x2119,
    DW_AT_GNU_deleted                       = 0x211a,
    DW_AT_GNU_dwo_name                      = 0x2130,
    DW_AT_GNU_dwo_id                        = 0x2131,
    DW_AT_GNU_ranges_base                   = 0x2132,
    DW_AT_GNU_addr_base                     = 0x2133,
    DW_AT_GNU_pubnames                      = 0x2134,
    DW_AT_GNU_pubtypes                      = 0x2135,
    DW_AT_GNU_discriminator                 = 0x2136,
    DW_AT_GNU_locviews                      = 0x2137,
    DW_AT_GNU_entry_view                    = 0x2138,

    DW_AT_LLVM_include_path  = 0x3e00,
    DW_AT_LLVM_config_macros = 0x3e01,
    DW_AT_LLVM_isysroot      = 0x3e02, # sysroot elsewhere
    DW_AT_LLVM_tag_offset    = 0x3e03,
    DW_AT_LLVM_apinotes      = 0x3e07,

    DW_AT_APPLE_optimized               = 0x3fe1,
    DW_AT_APPLE_flags                   = 0x3fe2,
    DW_AT_APPLE_isa                     = 0x3fe3,
    DW_AT_APPLE_block                   = 0x3fe4,
    DW_AT_APPLE_major_runtime_vers      = 0x3fe5,
    DW_AT_APPLE_runtime_class           = 0x3fe6,
    DW_AT_APPLE_omit_frame_ptr          = 0x3fe7,
    DW_AT_APPLE_property_name           = 0x3fe8,
    DW_AT_APPLE_property_getter         = 0x3fe9,
    DW_AT_APPLE_property_setter         = 0x3fea,
    DW_AT_APPLE_property_attribute      = 0x3feb,
    DW_AT_APPLE_objc_complete_type      = 0x3fec,
    DW_AT_APPLE_property                = 0x3fed,
    DW_AT_APPLE_objc_direct             = 0x3fee,
    DW_AT_APPLE_sdk                     = 0x3fef,

    _default_ = Pass,
)


ENUM_DW_FORM = dict(
    DW_FORM_null                = 0x00,
    DW_FORM_addr                = 0x01,
    DW_FORM_ref                 = 0x02,
    DW_FORM_block2              = 0x03,
    DW_FORM_block4              = 0x04,
    DW_FORM_data2               = 0x05,
    DW_FORM_data4               = 0x06,
    DW_FORM_data8               = 0x07,
    DW_FORM_string              = 0x08,
    DW_FORM_block               = 0x09,
    DW_FORM_block1              = 0x0a,
    DW_FORM_data1               = 0x0b,
    DW_FORM_flag                = 0x0c,
    DW_FORM_sdata               = 0x0d,
    DW_FORM_strp                = 0x0e,
    DW_FORM_udata               = 0x0f,
    DW_FORM_ref_addr            = 0x10,
    DW_FORM_ref1                = 0x11,
    DW_FORM_ref2                = 0x12,
    DW_FORM_ref4                = 0x13,
    DW_FORM_ref8                = 0x14,
    DW_FORM_ref_udata           = 0x15,
    DW_FORM_indirect            = 0x16,
    DW_FORM_sec_offset          = 0x17,
    DW_FORM_exprloc             = 0x18,
    DW_FORM_flag_present        = 0x19,
    DW_FORM_strx                = 0x1a,
    DW_FORM_addrx               = 0x1b,
    DW_FORM_ref_sup4            = 0x1c,
    DW_FORM_strp_sup            = 0x1d,
    DW_FORM_data16              = 0x1e,
    DW_FORM_line_strp           = 0x1f,
    DW_FORM_ref_sig8            = 0x20,
    DW_FORM_implicit_const      = 0x21,
    DW_FORM_loclistx            = 0x22,
    DW_FORM_rnglistx            = 0x23,
    DW_FORM_ref_sup8            = 0x24,
    DW_FORM_strx1               = 0x25,
    DW_FORM_strx2               = 0x26,
    DW_FORM_strx3               = 0x27,
    DW_FORM_strx4               = 0x28,
    DW_FORM_addrx1              = 0x29,
    DW_FORM_addrx2              = 0x2a,
    DW_FORM_addrx3              = 0x2b,
    DW_FORM_addrx4              = 0x2c,

    DW_FORM_GNU_addr_index      = 0x1f01,
    DW_FORM_GNU_str_index       = 0x1f02,
    DW_FORM_GNU_ref_alt         = 0x1f20,
    DW_FORM_GNU_strp_alt        = 0x1f21,
    _default_                   = Pass,
)

# Inverse mapping for ENUM_DW_FORM
DW_FORM_raw2name = dict((v, k) for k, v in ENUM_DW_FORM.items())

# See http://www.airs.com/blog/archives/460
DW_EH_encoding_flags = dict(
    DW_EH_PE_absptr   = 0x00,
    DW_EH_PE_uleb128  = 0x01,
    DW_EH_PE_udata2   = 0x02,
    DW_EH_PE_udata4   = 0x03,
    DW_EH_PE_udata8   = 0x04,

    DW_EH_PE_signed   = 0x08,
    DW_EH_PE_sleb128  = 0x09,
    DW_EH_PE_sdata2   = 0x0a,
    DW_EH_PE_sdata4   = 0x0b,
    DW_EH_PE_sdata8   = 0x0c,

    DW_EH_PE_pcrel    = 0x10,
    DW_EH_PE_textrel  = 0x20,
    DW_EH_PE_datarel  = 0x30,
    DW_EH_PE_funcrel  = 0x40,
    DW_EH_PE_aligned  = 0x50,
    DW_EH_PE_indirect = 0x80,

    DW_EH_PE_omit     = 0xff,
)

ENUM_DW_LNCT = dict(
    DW_LNCT_path             = 0x1,
    DW_LNCT_directory_index  = 0x2,
    DW_LNCT_timestamp        = 0x3,
    DW_LNCT_size             = 0x4,
    DW_LNCT_MD5              = 0x5,
    DW_LNCT_lo_user          = 0x2000,
    DW_LNCT_LLVM_source      = 0x2001,
    DW_LNCT_LLVM_is_MD5      = 0x2002,
    DW_LNCT_hi_user          = 0x3fff
)

ENUM_DW_UT = dict(
    DW_UT_compile       = 0x01,
    DW_UT_type          = 0x02,
    DW_UT_partial       = 0x03,
    DW_UT_skeleton      = 0x04,
    DW_UT_split_compile = 0x05,
    DW_UT_split_type    = 0x06,
    DW_UT_lo_user       = 0x80,
    DW_UT_hi_user       = 0xff
)

ENUM_DW_LLE = dict(
    DW_LLE_end_of_list      = 0x00,
    DW_LLE_base_addressx    = 0x01,
    DW_LLE_startx_endx      = 0x02,
    DW_LLE_startx_length    = 0x03,
    DW_LLE_offset_pair      = 0x04,
    DW_LLE_default_location = 0x05,
    DW_LLE_base_address     = 0x06,
    DW_LLE_start_end        = 0x07,
    DW_LLE_start_length     = 0x08
)

ENUM_DW_RLE = dict(
    DW_RLE_end_of_list   = 0x00,
    DW_RLE_base_addressx = 0x01,
    DW_RLE_startx_endx   = 0x02,
    DW_RLE_startx_length = 0x03,
    DW_RLE_offset_pair   = 0x04,
    DW_RLE_base_address  = 0x05,
    DW_RLE_start_end     = 0x06,
    DW_RLE_start_length  = 0x07
)

ENUM_DW_LANG = dict(
    DW_LANG_C89            = 0x0001,
    DW_LANG_C              = 0x0002,
    DW_LANG_Ada83          = 0x0003,
    DW_LANG_C_plus_plus    = 0x0004,
    DW_LANG_Cobol74        = 0x0005,
    DW_LANG_Cobol85        = 0x0006,
    DW_LANG_Fortran77      = 0x0007,
    DW_LANG_Fortran90      = 0x0008,
    DW_LANG_Pascal83       = 0x0009,
    DW_LANG_Modula2        = 0x000a,
    DW_LANG_Java           = 0x000b,
    DW_LANG_C99            = 0x000c,
    DW_LANG_Ada95          = 0x000d,
    DW_LANG_Fortran95      = 0x000e,
    DW_LANG_PLI            = 0x000f,
    DW_LANG_ObjC           = 0x0010,
    DW_LANG_ObjC_plus_plus = 0x0011,
    DW_LANG_UPC            = 0x0012,
    DW_LANG_D              = 0x0013,
    DW_LANG_Python         = 0x0014,
    DW_LANG_OpenCL         = 0x0015,
    DW_LANG_Go             = 0x0016,
    DW_LANG_Modula3        = 0x0017,
    DW_LANG_Haskell        = 0x0018,
    DW_LANG_C_plus_plus_03 = 0x0019,
    DW_LANG_C_plus_plus_11 = 0x001a,
    DW_LANG_OCaml          = 0x001b,
    DW_LANG_Rust           = 0x001c,
    DW_LANG_C11            = 0x001d,
    DW_LANG_Swift          = 0x001e,
    DW_LANG_Julia          = 0x001f,
    DW_LANG_Dylan          = 0x0020,
    DW_LANG_C_plus_plus_14 = 0x0021,
    DW_LANG_Fortran03      = 0x0022,
    DW_LANG_Fortran08      = 0x0023,
    DW_LANG_RenderScript   = 0x0024,
    DW_LANG_BLISS          = 0x0025,
    DW_LANG_lo_user        = 0x8000,
    DW_LANG_hi_user        = 0xffff
)

ENUM_DW_ATE = dict(
    DW_ATE_address         = 0x01,
    DW_ATE_boolean         = 0x02,
    DW_ATE_complex_float   = 0x03,
    DW_ATE_float           = 0x04,
    DW_ATE_signed          = 0x05,
    DW_ATE_signed_char     = 0x06,
    DW_ATE_unsigned        = 0x07,
    DW_ATE_unsigned_char   = 0x08,
    DW_ATE_imaginary_float = 0x09,
    DW_ATE_packed_decimal  = 0x0a,
    DW_ATE_numeric_string  = 0x0b,
    DW_ATE_edited          = 0x0c,
    DW_ATE_signed_fixed    = 0x0d,
    DW_ATE_unsigned_fixed  = 0x0e,
    DW_ATE_decimal_float   = 0x0f,
    DW_ATE_UTF             = 0x10,
    DW_ATE_UCS             = 0x11,
    DW_ATE_ASCII           = 0x12,
    DW_ATE_lo_user         = 0x80,
    DW_ATE_hi_user         = 0xff
)

ENUM_DW_ACCESS = dict(
    DW_ACCESS_public    = 0x01,
    DW_ACCESS_protected = 0x02,
    DW_ACCESS_private   = 0x03
)

ENUM_DW_INL = dict(
    DW_INL_not_inlined          = 0x00,
    DW_INL_inlined              = 0x01,
    DW_INL_declared_not_inlined = 0x02,
    DW_INL_declared_inlined     = 0x03
)

ENUM_DW_CC = dict(
    DW_CC_normal            = 0x01,
    DW_CC_program           = 0x02,
    DW_CC_nocall            = 0x03,
    DW_CC_pass_by_reference = 0x04,
    DW_CC_pass_by_value     = 0x05,
    DW_CC_lo_user           = 0x40,
    DW_CC_hi_user           = 0xff
)



#-------------------------------------------------------------------------------
# elftools example: dwarf_decode_address.py
#
# Decode an address in an ELF file to find out which function it belongs to
# and from which filename/line it comes in the original source file.
#
# Eli Bendersky (eliben@gmail.com)
# This code is in the public domain
#-------------------------------------------------------------------------------
from __future__ import print_function
import sys

# If pyelftools is not installed, the example can also run from the root or
# examples/ dir of the source distribution.
sys.path[0:0] = ['.', '..']

from elftools.common.utils import bytes2str
from elftools.dwarf.descriptions import describe_form_class
from elftools.elf.elffile import ELFFile


def process_file(filename, address):
    print('Processing file:', filename)
    with open(filename, 'rb') as f:
        elffile = ELFFile(f)

        if not elffile.has_dwarf_info():
            print('  file has no DWARF info')
            return

        # get_dwarf_info returns a DWARFInfo context object, which is the
        # starting point for all DWARF-based processing in pyelftools.
        dwarfinfo = elffile.get_dwarf_info()

        funcname = decode_funcname(dwarfinfo, address)
        file, line = decode_file_line(dwarfinfo, address)

        print('Function:', bytes2str(funcname))
        print('File:', bytes2str(file))
        print('Line:', line)


def decode_funcname(dwarfinfo, address):
    # Go over all DIEs in the DWARF information, looking for a subprogram
    # entry with an address range that includes the given address. Note that
    # this simplifies things by disregarding subprograms that may have
    # split address ranges.
    for CU in dwarfinfo.iter_CUs():
        for DIE in CU.iter_DIEs():
            try:
                if DIE.tag == 'DW_TAG_subprogram':
                    lowpc = DIE.attributes['DW_AT_low_pc'].value

                    # DWARF v4 in section 2.17 describes how to interpret the
                    # DW_AT_high_pc attribute based on the class of its form.
                    # For class 'address' it's taken as an absolute address
                    # (similarly to DW_AT_low_pc); for class 'constant', it's
                    # an offset from DW_AT_low_pc.
                    highpc_attr = DIE.attributes['DW_AT_high_pc']
                    highpc_attr_class = describe_form_class(highpc_attr.form)
                    if highpc_attr_class == 'address':
                        highpc = highpc_attr.value
                    elif highpc_attr_class == 'constant':
                        highpc = lowpc + highpc_attr.value
                    else:
                        print('Error: invalid DW_AT_high_pc class:',
                              highpc_attr_class)
                        continue

                    if lowpc <= address < highpc:
                        return DIE.attributes['DW_AT_name'].value
            except KeyError:
                continue
    return None


def decode_file_line(dwarfinfo, address):
    # Go over all the line programs in the DWARF information, looking for
    # one that describes the given address.
    for CU in dwarfinfo.iter_CUs():
        # First, look at line programs to find the file/line for the address
        lineprog = dwarfinfo.line_program_for_CU(CU)
        delta = 1 if lineprog.header.version < 5 else 0
        prevstate = None
        for entry in lineprog.get_entries():
            # We're interested in those entries where a new state is assigned
            if entry.state is None:
                continue
            # Looking for a range of addresses in two consecutive states that
            # contain the required address.
            if prevstate and prevstate.address <= address < entry.state.address:
                filename = lineprog['file_entry'][prevstate.file - delta].name
                line = prevstate.line
                return filename, line
            if entry.state.end_sequence:
                # For the state with `end_sequence`, `address` means the address
                # of the first byte after the target machine instruction
                # sequence and other information is meaningless. We clear
                # prevstate so that it's not used in the next iteration. Address
                # info is used in the above comparison to see if we need to use
                # the line information for the prevstate.
                prevstate = None
            else:
                prevstate = entry.state
    return None, None


if __name__ == '__main__':
    if sys.argv[1] == '--test':
        process_file(sys.argv[2], 0x400503)
        sys.exit(0)

    if len(sys.argv) < 3:
        print('Expected usage: {0} <address> <executable>'.format(sys.argv[0]))
        sys.exit(1)
    addr = int(sys.argv[1], 0)
    process_file(sys.argv[2], addr)



#-------------------------------------------------------------------------------
# elftools: dwarf/dwarfinfo.py
#
# DWARFInfo - Main class for accessing DWARF debug information
#
# Eli Bendersky (eliben@gmail.com)
# This code is in the public domain
#-------------------------------------------------------------------------------
import os
from collections import namedtuple, OrderedDict
from bisect import bisect_right

from ..construct.lib.container import Container
from ..common.exceptions import DWARFError
from ..common.utils import (struct_parse, dwarf_assert,
                            parse_cstring_from_stream)
from .structs import DWARFStructs
from .compileunit import CompileUnit
from .typeunit import TypeUnit
from .abbrevtable import AbbrevTable
from .lineprogram import LineProgram
from .callframe import CallFrameInfo
from .locationlists import LocationLists, LocationListsPair
from .ranges import RangeLists, RangeListsPair
from .aranges import ARanges
from .namelut import NameLUT
from .dwarf_util import _get_base_offset


# Describes a debug section
#
# stream: a stream object containing the data of this section
# name: section name in the container file
# global_offset: the global offset of the section in its container file
# size: the size of the section's data, in bytes
# address: the virtual address for the section's data
#
# 'name' and 'global_offset' are for descriptional purposes only and
# aren't strictly required for the DWARF parsing to work. 'address' is required
# to properly decode the special '.eh_frame' format.
#
DebugSectionDescriptor = namedtuple('DebugSectionDescriptor',
    'stream name global_offset size address')


# Some configuration parameters for the DWARF reader. This exists to allow
# DWARFInfo to be independent from any specific file format/container.
#
# little_endian:
#   boolean flag specifying whether the data in the file is little endian
#
# machine_arch:
#   Machine architecture as a string. For example 'x86' or 'x64'
#
# default_address_size:
#   The default address size for the container file (sizeof pointer, in bytes)
#
DwarfConfig = namedtuple('DwarfConfig',
    'little_endian machine_arch default_address_size')


class DWARFInfo(object):
    """ Acts also as a "context" to other major objects, bridging between
        various parts of the debug information.
    """
    def __init__(self,
            config,
            debug_info_sec,
            debug_aranges_sec,
            debug_abbrev_sec,
            debug_frame_sec,
            eh_frame_sec,
            debug_str_sec,
            debug_loc_sec,
            debug_ranges_sec,
            debug_line_sec,
            debug_pubtypes_sec,
            debug_pubnames_sec,
            debug_addr_sec,
            debug_str_offsets_sec,
            debug_line_str_sec,
            debug_loclists_sec,
            debug_rnglists_sec,
            debug_sup_sec,
            gnu_debugaltlink_sec,
            debug_types_sec
            ):
        """ config:
                A DwarfConfig object

            debug_*_sec:
                DebugSectionDescriptor for a section. Pass None for sections
                that don't exist. These arguments are best given with
                keyword syntax.
        """
        self.config = config
        self.debug_info_sec = debug_info_sec
        self.debug_aranges_sec = debug_aranges_sec
        self.debug_abbrev_sec = debug_abbrev_sec
        self.debug_frame_sec = debug_frame_sec
        self.eh_frame_sec = eh_frame_sec
        self.debug_str_sec = debug_str_sec
        self.debug_loc_sec = debug_loc_sec
        self.debug_ranges_sec = debug_ranges_sec
        self.debug_line_sec = debug_line_sec
        self.debug_addr_sec = debug_addr_sec
        self.debug_str_offsets_sec = debug_str_offsets_sec
        self.debug_line_str_sec = debug_line_str_sec
        self.debug_pubtypes_sec = debug_pubtypes_sec
        self.debug_pubnames_sec = debug_pubnames_sec
        self.debug_loclists_sec = debug_loclists_sec
        self.debug_rnglists_sec = debug_rnglists_sec
        self.debug_sup_sec = debug_sup_sec
        self.gnu_debugaltlink_sec = gnu_debugaltlink_sec
        self.debug_types_sec = debug_types_sec

        # Sets the supplementary_dwarfinfo to None. Client code can set this
        # to something else, typically a DWARFInfo file read from an ELFFile
        # which path is stored in the debug_sup_sec or gnu_debugaltlink_sec.
        self.supplementary_dwarfinfo = None

        # This is the DWARFStructs the context uses, so it doesn't depend on
        # DWARF format and address_size (these are determined per CU) - set them
        # to default values.
        self.structs = DWARFStructs(
            little_endian=self.config.little_endian,
            dwarf_format=32,
            address_size=self.config.default_address_size)

        # Cache for abbrev tables: a dict keyed by offset
        self._abbrevtable_cache = {}
        # Cache for program lines tables: a dict keyed by offset
        self._linetable_cache = {}
 
        # Cache of compile units and map of their offsets for bisect lookup.
        # Access with .iter_CUs(), .get_CU_containing(), and/or .get_CU_at().
        self._cu_cache = []
        self._cu_offsets_map = []

        # DWARF v4 type units by sig8 - OrderedDict created when needed
        self._type_units_by_sig = None

    @property
    def has_debug_info(self):
        """ Return whether this contains debug information.

        It can be not the case when the ELF only contains .eh_frame, which is
        encoded DWARF but not actually for debugging.
        """
        return bool(self.debug_info_sec)

    def has_debug_types(self):
        """ Return whether this contains debug types information.
        """
        return bool(self.debug_types_sec)

    def get_DIE_from_lut_entry(self, lut_entry):
        """ Get the DIE from the pubnames or putbtypes lookup table entry.

            lut_entry:
                A NameLUTEntry object from a NameLUT instance (see
                .get_pubmames and .get_pubtypes methods).
        """
        cu = self.get_CU_at(lut_entry.cu_ofs)
        return self.get_DIE_from_refaddr(lut_entry.die_ofs, cu)

    def get_DIE_from_refaddr(self, refaddr, cu=None):
        """ Given a .debug_info section offset of a DIE, return the DIE.

            refaddr:
                The refaddr may come from a DW_FORM_ref_addr attribute.

            cu:
                The compile unit object, if known.  If None a search
                from the closest offset less than refaddr will be performed.
        """
        if cu is None:
            cu = self.get_CU_containing(refaddr)
        return cu.get_DIE_from_refaddr(refaddr)
    
    def get_DIE_by_sig8(self, sig8):
        """ Find and return a DIE referenced by its type signature.
            sig8:
                The 8 byte signature (as a 64-bit unsigned integer)
            Returns the DIE with the given type signature by searching
            for the Type Unit with the matching signature then finding
            the DIE at the offset given by the type_die field in the
            Type Unit header.
                Signatures are an 64-bit unsigned integers computed by the
                DWARF producer as specified in the DWARF standard. Each
                Type Unit contains one signature and the offset to the
                corresponding DW_AT_type DIE in its unit header.
                Describing a type can generate several DIEs. By moving
                a DIE and its related DIEs to a Type Unit and generating
                a hash of the DIEs and attributes in a flattened form
                multiple Compile Units in a linked object can reference
                the same DIE in the overall DWARF structure.
            In DWARF v4 type units are identified by their appearance in the
            .debug_types section.
        """
        self._parse_debug_types()
        tu = self._type_units_by_sig.get(sig8)
        if tu is None:
            raise KeyError("Signature %016x not found in .debug_types" % sig8)
        return tu._get_cached_DIE(tu.tu_offset + tu['type_offset'])    

    def get_CU_containing(self, refaddr):
        """ Find the CU that includes the given reference address in the
            .debug_info section.

            refaddr:
                Either a refaddr of a DIE (possibly from a DW_FORM_ref_addr
                attribute) or the section offset of a CU (possibly from an
                aranges table).

           This function will parse and cache CUs until the search criteria
           is met, starting from the closest known offset lessthan or equal
           to the given address.
        """
        dwarf_assert(
            self.has_debug_info,
            'CU lookup but no debug info section')
        dwarf_assert(
            0 <= refaddr < self.debug_info_sec.size,
            "refaddr %s beyond .debug_info size" % refaddr)

        # The CU containing the DIE we desire will be to the right of the
        # DIE insert point.  If we have a CU address, then it will be a
        # match but the right insert minus one will still be the item.
        # The first CU starts at offset 0, so start there if cache is empty.
        i = bisect_right(self._cu_offsets_map, refaddr)
        start = self._cu_offsets_map[i - 1] if i > 0 else 0

        # parse CUs until we find one containing the desired address
        for cu in self._parse_CUs_iter(start):
            if cu.cu_offset <= refaddr < cu.cu_offset + cu.size:
                return cu

        raise ValueError("CU for reference address %s not found" % refaddr)

    def get_CU_at(self, offset):
        """ Given a CU header offset, return the parsed CU.

            offset:
                The offset may be from an accelerated access table such as
                the public names, public types, address range table, or
                prior use.

            This function will directly parse the CU doing no validation of
            the offset beyond checking the size of the .debug_info section.
        """
        dwarf_assert(
            self.has_debug_info,
            'CU lookup but no debug info section')
        dwarf_assert(
            0 <= offset < self.debug_info_sec.size,
            "offset %s beyond .debug_info size" % offset)

        return self._cached_CU_at_offset(offset)

    def get_TU_by_sig8(self, sig8):
        """ Find and return a Type Unit referenced by its signature

            sig8:
                The 8 byte unique signature (as a 64-bit unsigned integer)

            Returns the TU with the given type signature by parsing the
            .debug_types section.

        """
        self._parse_debug_types()
        tu = self._type_units_by_sig.get(sig8)
        if tu is None:
            raise KeyError("Signature %016x not found in .debug_types" % sig8)
        return tu

    def iter_CUs(self):
        """ Yield all the compile units (CompileUnit objects) in the debug info
        """
        return self._parse_CUs_iter()

    def iter_TUs(self):
        """Yield all the type units (TypeUnit objects) in the debug_types
        """
        return self._parse_TUs_iter()

    def get_abbrev_table(self, offset):
        """ Get an AbbrevTable from the given offset in the debug_abbrev
            section.

            The only verification done on the offset is that it's within the
            bounds of the section (if not, an exception is raised).
            It is the caller's responsibility to make sure the offset actually
            points to a valid abbreviation table.

            AbbrevTable objects are cached internally (two calls for the same
            offset will return the same object).
        """
        dwarf_assert(
            offset < self.debug_abbrev_sec.size,
            "Offset '0x%x' to abbrev table out of section bounds" % offset)
        if offset not in self._abbrevtable_cache:
            self._abbrevtable_cache[offset] = AbbrevTable(
                structs=self.structs,
                stream=self.debug_abbrev_sec.stream,
                offset=offset)
        return self._abbrevtable_cache[offset]

    def get_string_from_table(self, offset):
        """ Obtain a string from the string table section, given an offset
            relative to the section.
        """
        return parse_cstring_from_stream(self.debug_str_sec.stream, offset)

    def get_string_from_linetable(self, offset):
        """ Obtain a string from the string table section, given an offset
            relative to the section.
        """
        return parse_cstring_from_stream(self.debug_line_str_sec.stream, offset)

    def line_program_for_CU(self, CU):
        """ Given a CU object, fetch the line program it points to from the
            .debug_line section.
            If the CU doesn't point to a line program, return None.

            Note about directory and file names. They are returned as two collections
            in the lineprogram object's header - include_directory and file_entry.

            In DWARFv5, they have introduced a different, extensible format for those
            collections. So in a lineprogram v5+, there are two more collections in
            the header - directories and file_names. Those might contain extra DWARFv5
            information that is not exposed in include_directory and file_entry.
        """
        # The line program is pointed to by the DW_AT_stmt_list attribute of
        # the top DIE of a CU.
        top_DIE = CU.get_top_DIE()
        if 'DW_AT_stmt_list' in top_DIE.attributes:
            return self._parse_line_program_at_offset(
                    top_DIE.attributes['DW_AT_stmt_list'].value, CU.structs)
        else:
            return None

    def has_CFI(self):
        """ Does this dwarf info have a dwarf_frame CFI section?
        """
        return self.debug_frame_sec is not None

    def CFI_entries(self):
        """ Get a list of dwarf_frame CFI entries from the .debug_frame section.
        """
        cfi = CallFrameInfo(
            stream=self.debug_frame_sec.stream,
            size=self.debug_frame_sec.size,
            address=self.debug_frame_sec.address,
            base_structs=self.structs)
        return cfi.get_entries()

    def has_EH_CFI(self):
        """ Does this dwarf info have a eh_frame CFI section?
        """
        return self.eh_frame_sec is not None

    def EH_CFI_entries(self):
        """ Get a list of eh_frame CFI entries from the .eh_frame section.
        """
        cfi = CallFrameInfo(
            stream=self.eh_frame_sec.stream,
            size=self.eh_frame_sec.size,
            address=self.eh_frame_sec.address,
            base_structs=self.structs,
            for_eh_frame=True)
        return cfi.get_entries()

    def get_pubtypes(self):
        """
        Returns a NameLUT object that contains information read from the
        .debug_pubtypes section in the ELF file.

        NameLUT is essentially a dictionary containing the CU/DIE offsets of
        each symbol. See the NameLUT doc string for more details.
        """

        if self.debug_pubtypes_sec:
            return NameLUT(self.debug_pubtypes_sec.stream,
                    self.debug_pubtypes_sec.size,
                    self.structs)
        else:
            return None

    def get_pubnames(self):
        """
        Returns a NameLUT object that contains information read from the
        .debug_pubnames section in the ELF file.

        NameLUT is essentially a dictionary containing the CU/DIE offsets of
        each symbol. See the NameLUT doc string for more details.
        """

        if self.debug_pubnames_sec:
            return NameLUT(self.debug_pubnames_sec.stream,
                    self.debug_pubnames_sec.size,
                    self.structs)
        else:
            return None

    def get_aranges(self):
        """ Get an ARanges object representing the .debug_aranges section of
            the DWARF data, or None if the section doesn't exist
        """
        if self.debug_aranges_sec:
            return ARanges(self.debug_aranges_sec.stream,
                self.debug_aranges_sec.size,
                self.structs)
        else:
            return None

    def location_lists(self):
        """ Get a LocationLists object representing the .debug_loc/debug_loclists section of
            the DWARF data, or None if this section doesn't exist.

            If both sections exist, it returns a LocationListsPair.
        """
        if self.debug_loclists_sec and self.debug_loc_sec is None:
            return LocationLists(self.debug_loclists_sec.stream, self.structs, 5, self)
        elif self.debug_loc_sec and self.debug_loclists_sec is None:
            return LocationLists(self.debug_loc_sec.stream, self.structs, 4, self)
        elif self.debug_loc_sec and self.debug_loclists_sec:
            return LocationListsPair(self.debug_loc_sec.stream, self.debug_loclists_sec.stream, self.structs, self)
        else:
            return None

    def range_lists(self):
        """ Get a RangeLists object representing the .debug_ranges/.debug_rnglists section of
            the DWARF data, or None if this section doesn't exist.

            If both sections exist, it returns a RangeListsPair.
        """
        if self.debug_rnglists_sec and self.debug_ranges_sec is None:
            return RangeLists(self.debug_rnglists_sec.stream, self.structs, 5, self)
        elif self.debug_ranges_sec and self.debug_rnglists_sec is None:
            return RangeLists(self.debug_ranges_sec.stream, self.structs, 4, self)
        elif self.debug_ranges_sec and self.debug_rnglists_sec:
            return RangeListsPair(self.debug_ranges_sec.stream, self.debug_rnglists_sec.stream, self.structs, self)
        else:
            return None

    def get_addr(self, cu, addr_index):
        """Provided a CU and an index, retrieves an address from the debug_addr section
        """
        if not self.debug_addr_sec:
            raise DWARFError('The file does not contain a debug_addr section for indirect address access')
        # Selectors are not supported, but no assert on that. TODO?
        cu_addr_base = _get_base_offset(cu, 'DW_AT_addr_base')
        return struct_parse(cu.structs.the_Dwarf_target_addr, self.debug_addr_sec.stream, cu_addr_base + addr_index*cu.header.address_size)            

    #------ PRIVATE ------#

    def _parse_CUs_iter(self, offset=0):
        """ Iterate CU objects in order of appearance in the debug_info section.

            offset:
                The offset of the first CU to yield.  Additional iterations
                will return the sequential unit objects.

            See .iter_CUs(), .get_CU_containing(), and .get_CU_at().
        """
        if self.debug_info_sec is None:
            return

        while offset < self.debug_info_sec.size:
            cu = self._cached_CU_at_offset(offset)
            # Compute the offset of the next CU in the section. The unit_length
            # field of the CU header contains its size not including the length
            # field itself.
            offset = (offset +
                      cu['unit_length'] +
                      cu.structs.initial_length_field_size())
            yield cu

    def _parse_TUs_iter(self, offset=0):
        """ Iterate Type Unit objects in order of appearance in the debug_types section.

            offset:
                The offset of the first TU to yield.  Additional iterations
                will return the sequential unit objects.

                See .iter_TUs().
        """
        if self.debug_types_sec is None:
            return

        while offset < self.debug_types_sec.size:
            tu = self._parse_TU_at_offset(offset)
            # Compute the offset of the next TU in the section. The unit_length
            # field of the TU header contains its size not including the length
            # field itself.
            offset = (offset +
                      tu['unit_length'] +
                      tu.structs.initial_length_field_size())

            yield tu

    def _parse_debug_types(self):
        """ Check if the .debug_types section is previously parsed. If not,
            parse all TUs and store them in an OrderedDict using their unique
            64-bit signature as the key.

            See .get_TU_by_sig8().
        """
        if self._type_units_by_sig is not None:
            return
        self._type_units_by_sig = OrderedDict()

        if self.debug_types_sec is None:
            return

        # Parse all the Type Units in the types section for access by sig8
        offset = 0
        while offset < self.debug_types_sec.size:
            tu = self._parse_TU_at_offset(offset)
            # Compute the offset of the next TU in the section. The unit_length
            # field of the TU header contains its size not including the length
            # field itself.
            offset = (offset +
                      tu['unit_length'] +
                      tu.structs.initial_length_field_size())
            self._type_units_by_sig[tu['signature']] = tu

    def _cached_CU_at_offset(self, offset):
        """ Return the CU with unit header at the given offset into the
            debug_info section from the cache.  If not present, the unit is
            header is parsed and the object is installed in the cache.

            offset:
                The offset of the unit header in the .debug_info section
                to of the unit to fetch from the cache.

            See get_CU_at().
        """
        # Find the insert point for the requested offset.  With bisect_right,
        # if this entry is present in the cache it will be the prior entry.
        i = bisect_right(self._cu_offsets_map, offset)
        if i >= 1 and offset == self._cu_offsets_map[i - 1]:
            return self._cu_cache[i - 1]

        # Parse the CU and insert the offset and object into the cache.
        # The ._cu_offsets_map[] contains just the numeric offsets for the
        # bisect_right search while the parallel indexed ._cu_cache[] holds
        # the object references.
        cu = self._parse_CU_at_offset(offset)
        self._cu_offsets_map.insert(i, offset)
        self._cu_cache.insert(i, cu)
        return cu

    def _parse_CU_at_offset(self, offset):
        """ Parse and return a CU at the given offset in the debug_info stream.
        """
        # Section 7.4 (32-bit and 64-bit DWARF Formats) of the DWARF spec v3
        # states that the first 32-bit word of the CU header determines
        # whether the CU is represented with 32-bit or 64-bit DWARF format.
        #
        # So we peek at the first word in the CU header to determine its
        # dwarf format. Based on it, we then create a new DWARFStructs
        # instance suitable for this CU and use it to parse the rest.
        #
        initial_length = struct_parse(
            self.structs.the_Dwarf_uint32, self.debug_info_sec.stream, offset)
        dwarf_format = 64 if initial_length == 0xFFFFFFFF else 32


        # Temporary structs for parsing the header
        # The structs for the rest of the CU depend on the header data.
        #
        cu_structs = DWARFStructs(
            little_endian=self.config.little_endian,
            dwarf_format=dwarf_format,
            address_size=4,
            dwarf_version=2)

        cu_header = struct_parse(
            cu_structs.Dwarf_CU_header, self.debug_info_sec.stream, offset)

        # structs for the rest of the CU, taking into account bitness and DWARF version
        cu_structs = DWARFStructs(
            little_endian=self.config.little_endian,
            dwarf_format=dwarf_format,
            address_size=cu_header['address_size'],
            dwarf_version=cu_header['version'])

        cu_die_offset = self.debug_info_sec.stream.tell()
        dwarf_assert(
            self._is_supported_version(cu_header['version']),
            "Expected supported DWARF version. Got '%s'" % cu_header['version'])
        return CompileUnit(
                header=cu_header,
                dwarfinfo=self,
                structs=cu_structs,
                cu_offset=offset,
                cu_die_offset=cu_die_offset)

    def _parse_TU_at_offset(self, offset):
        """ Parse and return a Type Unit (TU) at the given offset in the debug_types stream.
        """
        # Section 7.4 (32-bit and 64-bit DWARF Formats) of the DWARF spec v4
        # states that the first 32-bit word of the TU header determines
        # whether the TU is represented with 32-bit or 64-bit DWARF format.
        #
        # So we peek at the first word in the TU header to determine its
        # dwarf format. Based on it, we then create a new DWARFStructs
        # instance suitable for this TU and use it to parse the rest.
        #
        initial_length = struct_parse(
            self.structs.the_Dwarf_uint32, self.debug_types_sec.stream, offset)
        dwarf_format = 64 if initial_length == 0xFFFFFFFF else 32

        # Temporary structs for parsing the header
        # The structs for the rest of the TU depend on the header data.
        #
        tu_structs = DWARFStructs(
            little_endian=self.config.little_endian,
            dwarf_format=dwarf_format,
            address_size=4,
            dwarf_version=2)

        tu_header = struct_parse(
            tu_structs.Dwarf_TU_header, self.debug_types_sec.stream, offset)

        # structs for the rest of the TU, taking into account bit-width and DWARF version
        tu_structs = DWARFStructs(
            little_endian=self.config.little_endian,
            dwarf_format=dwarf_format,
            address_size=tu_header['address_size'],
            dwarf_version=tu_header['version'])

        tu_die_offset = self.debug_types_sec.stream.tell()
        dwarf_assert(
            self._is_supported_version(tu_header['version']),
            "Expected supported DWARF version. Got '%s'" % tu_header['version'])
        return TypeUnit(
            header=tu_header,
            dwarfinfo=self,
            structs=tu_structs,
            tu_offset=offset,
            tu_die_offset=tu_die_offset)

    def _is_supported_version(self, version):
        """ DWARF version supported by this parser
        """
        return 2 <= version <= 5

    def _parse_line_program_at_offset(self, offset, structs):
        """ Given an offset to the .debug_line section, parse the line program
            starting at this offset in the section and return it.
            structs is the DWARFStructs object used to do this parsing.
        """

        if offset in self._linetable_cache:
            return self._linetable_cache[offset]

        lineprog_header = struct_parse(
            structs.Dwarf_lineprog_header,
            self.debug_line_sec.stream,
            offset)

        # DWARF5: resolve names
        def resolve_strings(self, lineprog_header, format_field, data_field):
            if lineprog_header.get(format_field, False):
                data = lineprog_header[data_field]
                for field in lineprog_header[format_field]:
                    def replace_value(data, content_type, replacer):
                        for entry in data:
                            entry[content_type] = replacer(entry[content_type])

                    if field.form == 'DW_FORM_line_strp':
                        replace_value(data, field.content_type, self.get_string_from_linetable)
                    elif field.form == 'DW_FORM_strp':
                        replace_value(data, field.content_type, self.get_string_from_table)
                    elif field.form in ('DW_FORM_strp_sup', 'DW_FORM_GNU_strp_alt'):
                        if self.supplementary_dwarfinfo:
                            replace_value(data, field.content_type, self.supplementary_dwarfinfo.get_string_fromtable)
                        else:
                            replace_value(data, field.content_type, lambda x: str(x))
                    elif field.form in ('DW_FORM_strp_sup', 'DW_FORM_strx', 'DW_FORM_strx1', 'DW_FORM_strx2', 'DW_FORM_strx3', 'DW_FORM_strx4'):
                        raise NotImplementedError()

        resolve_strings(self, lineprog_header, 'directory_entry_format', 'directories')
        resolve_strings(self, lineprog_header, 'file_name_entry_format', 'file_names')

        # DWARF5: provide compatible file/directory name arrays for legacy lineprogram consumers
        if lineprog_header.get('directories', False):
            lineprog_header.include_directory = tuple(d.DW_LNCT_path for d in lineprog_header.directories)
        if lineprog_header.get('file_names', False):
            lineprog_header.file_entry = tuple(
                Container(**{
                    'name':e.get('DW_LNCT_path'),
                    'dir_index': e.get('DW_LNCT_directory_index'),
                    'mtime': e.get('DW_LNCT_timestamp'),
                    'length': e.get('DW_LNCT_size')})
                for e in lineprog_header.file_names)

        # Calculate the offset to the next line program (see DWARF 6.2.4)
        end_offset = (  offset + lineprog_header['unit_length'] +
                        structs.initial_length_field_size())

        lineprogram = LineProgram(
            header=lineprog_header,
            stream=self.debug_line_sec.stream,
            structs=structs,
            program_start_offset=self.debug_line_sec.stream.tell(),
            program_end_offset=end_offset)

        self._linetable_cache[offset] = lineprogram
        return lineprogram

    def parse_debugsupinfo(self):
        """
        Extract a filename from .debug_sup, .gnu_debualtlink sections.
        """
        if self.debug_sup_sec is not None:
            self.debug_sup_sec.stream.seek(0)
            suplink = self.structs.Dwarf_debugsup.parse_stream(self.debug_sup_sec.stream)
            if suplink.is_supplementary == 0:
                return suplink.sup_filename
        if self.gnu_debugaltlink_sec is not None:
            self.gnu_debugaltlink_sec.stream.seek(0)
            suplink = self.structs.Dwarf_debugaltlink.parse_stream(self.gnu_debugaltlink_sec.stream)
            return suplink.sup_filename
        # The section .gnu_debuglink with similarly looking contents
        # has a different meaning - it doesn't point at supplementary DWARF,
        # which is meant to be referenced from primary DWARF,
        # it points at DWARF proper.
        return None




# -------------------------------------------------------------------------------
# elftools: ehabi/decoder.py
#
# Decode ARM exception handler bytecode.
#
# LeadroyaL (leadroyal@qq.com)
# This code is in the public domain
# -------------------------------------------------------------------------------
from collections import namedtuple


class EHABIBytecodeDecoder(object):
    """ Decoder of a sequence of ARM exception handler abi bytecode.

        Reference:
        https://github.com/llvm/llvm-project/blob/master/llvm/tools/llvm-readobj/ARMEHABIPrinter.h
        https://developer.arm.com/documentation/ihi0038/b/

        Accessible attributes:

            mnemonic_array:
                MnemonicItem array.

        Parameters:

            bytecode_array:
                Integer array, raw data of bytecode.

    """

    def __init__(self, bytecode_array):
        self._bytecode_array = bytecode_array
        self._index = None
        self.mnemonic_array = None
        self._decode()

    def _decode(self):
        """ Decode bytecode array, put result into mnemonic_array.
        """
        self._index = 0
        self.mnemonic_array = []
        while self._index < len(self._bytecode_array):
            for mask, value, handler in self.ring:
                if (self._bytecode_array[self._index] & mask) == value:
                    start_idx = self._index
                    mnemonic = handler(self)
                    end_idx = self._index
                    self.mnemonic_array.append(
                        MnemonicItem(self._bytecode_array[start_idx: end_idx], mnemonic))
                    break

    def _decode_00xxxxxx(self):
        #   SW.startLine() << format("0x%02X      ; vsp = vsp + %u\n", Opcode,
        #                            ((Opcode & 0x3f) << 2) + 4);
        opcode = self._bytecode_array[self._index]
        self._index += 1
        return 'vsp = vsp + %u' % (((opcode & 0x3f) << 2) + 4)

    def _decode_01xxxxxx(self):
        # SW.startLine() << format("0x%02X      ; vsp = vsp - %u\n", Opcode,
        #                          ((Opcode & 0x3f) << 2) + 4);
        opcode = self._bytecode_array[self._index]
        self._index += 1
        return 'vsp = vsp - %u' % (((opcode & 0x3f) << 2) + 4)

    gpr_register_names = ("r0", "r1", "r2", "r3", "r4", "r5", "r6", "r7",
                          "r8", "r9", "r10", "fp", "ip", "sp", "lr", "pc")

    def _calculate_range(self, start, count):
        return ((1 << (count + 1)) - 1) << start

    def _printGPR(self, gpr_mask):
        hits = [self.gpr_register_names[i] for i in range(32) if gpr_mask & (1 << i) != 0]
        return '{%s}' % ', '.join(hits)

    def _print_registers(self, vfp_mask, prefix):
        hits = [prefix + str(i) for i in range(32) if vfp_mask & (1 << i) != 0]
        return '{%s}' % ', '.join(hits)

    def _decode_1000iiii_iiiiiiii(self):
        op0 = self._bytecode_array[self._index]
        self._index += 1
        op1 = self._bytecode_array[self._index]
        self._index += 1
        #   uint16_t GPRMask = (Opcode1 << 4) | ((Opcode0 & 0x0f) << 12);
        #   SW.startLine()
        #     << format("0x%02X 0x%02X ; %s",
        #               Opcode0, Opcode1, GPRMask ? "pop " : "refuse to unwind");
        #   if (GPRMask)
        #     PrintGPR(GPRMask);
        gpr_mask = (op1 << 4) | ((op0 & 0x0f) << 12)
        if gpr_mask == 0:
            return 'refuse to unwind'
        else:
            return 'pop %s' % self._printGPR(gpr_mask)

    def _decode_10011101(self):
        self._index += 1
        return 'reserved (ARM MOVrr)'

    def _decode_10011111(self):
        self._index += 1
        return 'reserved (WiMMX MOVrr)'

    def _decode_1001nnnn(self):
        # SW.startLine() << format("0x%02X      ; vsp = r%u\n", Opcode, (Opcode & 0x0f));
        opcode = self._bytecode_array[self._index]
        self._index += 1
        return 'vsp = r%u' % (opcode & 0x0f)

    def _decode_10100nnn(self):
        # SW.startLine() << format("0x%02X      ; pop ", Opcode);
        # PrintGPR((((1 << ((Opcode & 0x7) + 1)) - 1) << 4));
        opcode = self._bytecode_array[self._index]
        self._index += 1
        return 'pop %s' % self._printGPR(self._calculate_range(4, opcode & 0x07))

    def _decode_10101nnn(self):
        # SW.startLine() << format("0x%02X      ; pop ", Opcode);
        # PrintGPR((((1 << ((Opcode & 0x7) + 1)) - 1) << 4) | (1 << 14));
        opcode = self._bytecode_array[self._index]
        self._index += 1
        return 'pop %s' % self._printGPR(self._calculate_range(4, opcode & 0x07) | (1 << 14))

    def _decode_10110000(self):
        # SW.startLine() << format("0x%02X      ; finish\n", Opcode);
        self._index += 1
        return 'finish'

    def _decode_10110001_0000iiii(self):
        # SW.startLine()
        #   << format("0x%02X 0x%02X ; %s", Opcode0, Opcode1,
        #             ((Opcode1 & 0xf0) || Opcode1 == 0x00) ? "spare" : "pop ");
        # if (((Opcode1 & 0xf0) == 0x00) && Opcode1)
        #   PrintGPR((Opcode1 & 0x0f));
        self._index += 1  # skip constant byte
        op1 = self._bytecode_array[self._index]
        self._index += 1
        if (op1 & 0xf0) != 0 or op1 == 0x00:
            return 'spare'
        else:
            return 'pop %s' % self._printGPR((op1 & 0x0f))

    def _decode_10110010_uleb128(self):
        #  SmallVector<uint8_t, 4> ULEB;
        #  do { ULEB.push_back(Opcodes[OI ^ 3]); } while (Opcodes[OI++ ^ 3] & 0x80);
        #  uint64_t Value = 0;
        #  for (unsigned BI = 0, BE = ULEB.size(); BI != BE; ++BI)
        #    Value = Value | ((ULEB[BI] & 0x7f) << (7 * BI));
        #  OS << format("; vsp = vsp + %" PRIu64 "\n", 0x204 + (Value << 2));
        self._index += 1   # skip constant byte
        uleb_buffer = [self._bytecode_array[self._index]]
        self._index += 1
        while self._bytecode_array[self._index] & 0x80 == 0:
            uleb_buffer.append(self._bytecode_array[self._index])
            self._index += 1
        value = 0
        for b in reversed(uleb_buffer):
            value = (value << 7) + (b & 0x7F)
        return 'vsp = vsp + %u' % (0x204 + (value << 2))

    def _decode_10110011_sssscccc(self):
        # these two decoders are equal
        return self._decode_11001001_sssscccc()

    def _decode_101101nn(self):
        return self._spare()

    def _decode_10111nnn(self):
        #  SW.startLine() << format("0x%02X      ; pop ", Opcode);
        #  PrintRegisters((((1 << ((Opcode & 0x07) + 1)) - 1) << 8), "d");
        opcode = self._bytecode_array[self._index]
        self._index += 1
        return 'pop %s' % self._print_registers(self._calculate_range(8, opcode & 0x07), "d")

    def _decode_11000110_sssscccc(self):
        #  SW.startLine() << format("0x%02X 0x%02X ; pop ", Opcode0, Opcode1);
        #  uint8_t Start = ((Opcode1 & 0xf0) >> 4);
        #  uint8_t Count = ((Opcode1 & 0x0f) >> 0);
        #  PrintRegisters((((1 << (Count + 1)) - 1) << Start), "wR");
        self._index += 1  # skip constant byte
        op1 = self._bytecode_array[self._index]
        self._index += 1
        start = ((op1 & 0xf0) >> 4)
        count = ((op1 & 0x0f) >> 0)
        return 'pop %s' % self._print_registers(self._calculate_range(start, count), "wR")

    def _decode_11000111_0000iiii(self):
        #   SW.startLine()
        #     << format("0x%02X 0x%02X ; %s", Opcode0, Opcode1,
        #               ((Opcode1 & 0xf0) || Opcode1 == 0x00) ? "spare" : "pop ");
        #   if ((Opcode1 & 0xf0) == 0x00 && Opcode1)
        #       PrintRegisters(Opcode1 & 0x0f, "wCGR");
        self._index += 1  # skip constant byte
        op1 = self._bytecode_array[self._index]
        self._index += 1
        if (op1 & 0xf0) != 0 or op1 == 0x00:
            return 'spare'
        else:
            return 'pop %s' % self._print_registers(op1 & 0x0f, "wCGR")

    def _decode_11001000_sssscccc(self):
        #   SW.startLine() << format("0x%02X 0x%02X ; pop ", Opcode0, Opcode1);
        #   uint8_t Start = 16 + ((Opcode1 & 0xf0) >> 4);
        #   uint8_t Count = ((Opcode1 & 0x0f) >> 0);
        #   PrintRegisters((((1 << (Count + 1)) - 1) << Start), "d");
        self._index += 1  # skip constant byte
        op1 = self._bytecode_array[self._index]
        self._index += 1
        start = 16 + ((op1 & 0xf0) >> 4)
        count = ((op1 & 0x0f) >> 0)
        return 'pop %s' % self._print_registers(self._calculate_range(start, count), "d")

    def _decode_11001001_sssscccc(self):
        #   SW.startLine() << format("0x%02X 0x%02X ; pop ", Opcode0, Opcode1);
        #   uint8_t Start = ((Opcode1 & 0xf0) >> 4);
        #   uint8_t Count = ((Opcode1 & 0x0f) >> 0);
        #   PrintRegisters((((1 << (Count + 1)) - 1) << Start), "d");
        self._index += 1  # skip constant byte
        op1 = self._bytecode_array[self._index]
        self._index += 1
        start = ((op1 & 0xf0) >> 4)
        count = ((op1 & 0x0f) >> 0)
        return 'pop %s' % self._print_registers(self._calculate_range(start, count), "d")

    def _decode_11001yyy(self):
        return self._spare()

    def _decode_11000nnn(self):
        #   SW.startLine() << format("0x%02X      ; pop ", Opcode);
        #   PrintRegisters((((1 << ((Opcode & 0x07) + 1)) - 1) << 10), "wR");
        opcode = self._bytecode_array[self._index]
        self._index += 1
        return 'pop %s' % self._print_registers(self._calculate_range(10, opcode & 0x07), "wR")

    def _decode_11010nnn(self):
        # these two decoders are equal
        return self._decode_10111nnn()

    def _decode_11xxxyyy(self):
        return self._spare()

    def _spare(self):
        self._index += 1
        return 'spare'

    _DECODE_RECIPE_TYPE = namedtuple('_DECODE_RECIPE_TYPE', 'mask value handler')

    ring = (
        _DECODE_RECIPE_TYPE(mask=0xc0, value=0x00, handler=_decode_00xxxxxx),
        _DECODE_RECIPE_TYPE(mask=0xc0, value=0x40, handler=_decode_01xxxxxx),
        _DECODE_RECIPE_TYPE(mask=0xf0, value=0x80, handler=_decode_1000iiii_iiiiiiii),
        _DECODE_RECIPE_TYPE(mask=0xff, value=0x9d, handler=_decode_10011101),
        _DECODE_RECIPE_TYPE(mask=0xff, value=0x9f, handler=_decode_10011111),
        _DECODE_RECIPE_TYPE(mask=0xf0, value=0x90, handler=_decode_1001nnnn),
        _DECODE_RECIPE_TYPE(mask=0xf8, value=0xa0, handler=_decode_10100nnn),
        _DECODE_RECIPE_TYPE(mask=0xf8, value=0xa8, handler=_decode_10101nnn),
        _DECODE_RECIPE_TYPE(mask=0xff, value=0xb0, handler=_decode_10110000),
        _DECODE_RECIPE_TYPE(mask=0xff, value=0xb1, handler=_decode_10110001_0000iiii),
        _DECODE_RECIPE_TYPE(mask=0xff, value=0xb2, handler=_decode_10110010_uleb128),
        _DECODE_RECIPE_TYPE(mask=0xff, value=0xb3, handler=_decode_10110011_sssscccc),
        _DECODE_RECIPE_TYPE(mask=0xfc, value=0xb4, handler=_decode_101101nn),
        _DECODE_RECIPE_TYPE(mask=0xf8, value=0xb8, handler=_decode_10111nnn),
        _DECODE_RECIPE_TYPE(mask=0xff, value=0xc6, handler=_decode_11000110_sssscccc),
        _DECODE_RECIPE_TYPE(mask=0xff, value=0xc7, handler=_decode_11000111_0000iiii),
        _DECODE_RECIPE_TYPE(mask=0xff, value=0xc8, handler=_decode_11001000_sssscccc),
        _DECODE_RECIPE_TYPE(mask=0xff, value=0xc9, handler=_decode_11001001_sssscccc),
        _DECODE_RECIPE_TYPE(mask=0xc8, value=0xc8, handler=_decode_11001yyy),
        _DECODE_RECIPE_TYPE(mask=0xf8, value=0xc0, handler=_decode_11000nnn),
        _DECODE_RECIPE_TYPE(mask=0xf8, value=0xd0, handler=_decode_11010nnn),
        _DECODE_RECIPE_TYPE(mask=0xc0, value=0xc0, handler=_decode_11xxxyyy),
    )


class MnemonicItem(object):
    """ Single mnemonic item.
    """

    def __init__(self, bytecode, mnemonic):
        self.bytecode = bytecode
        self.mnemonic = mnemonic

    def __repr__(self):
        return '%s ; %s' % (' '.join(['0x%02x' % x for x in self.bytecode]), self.mnemonic)



#-------------------------------------------------------------------------------
# elftools example: dwarf_location_info.py
#
# Examine DIE entries which have either location list values or location
# expression values and decode that information.
#
# Location information can either be completely contained within a DIE
# (using 'DW_FORM_exprloc' in DWARFv4 or 'DW_FORM_block1' in earlier
# versions) or be a reference to a location list contained within
# the .debug_loc section (using 'DW_FORM_sec_offset' in DWARFv4 or
# 'DW_FORM_data4' / 'DW_FORM_data8' in earlier versions).
#
# The LocationParser object parses the DIE attributes and handles both
# formats.
#
# The directory 'test/testfiles_for_location_info' contains test files with
# location information represented in both DWARFv4 and DWARFv2 forms.
#
# Eli Bendersky (eliben@gmail.com)
# This code is in the public domain
#-------------------------------------------------------------------------------
from __future__ import print_function
import sys

# If pyelftools is not installed, the example can also run from the root or
# examples/ dir of the source distribution.
sys.path[0:0] = ['.', '..']

from elftools.elf.elffile import ELFFile
from elftools.dwarf.descriptions import (
    describe_DWARF_expr, set_global_machine_arch)
from elftools.dwarf.locationlists import (
    LocationEntry, LocationExpr, LocationParser)

def process_file(filename):
    print('Processing file:', filename)
    with open(filename, 'rb') as f:
        elffile = ELFFile(f)

        if not elffile.has_dwarf_info():
            print('  file has no DWARF info')
            return

        # get_dwarf_info returns a DWARFInfo context object, which is the
        # starting point for all DWARF-based processing in pyelftools.
        dwarfinfo = elffile.get_dwarf_info()

        # The location lists are extracted by DWARFInfo from the .debug_loc
        # section, and returned here as a LocationLists object.
        location_lists = dwarfinfo.location_lists()

        # This is required for the descriptions module to correctly decode
        # register names contained in DWARF expressions.
        set_global_machine_arch(elffile.get_machine_arch())

        # Create a LocationParser object that parses the DIE attributes and
        # creates objects representing the actual location information.
        loc_parser = LocationParser(location_lists)

        for CU in dwarfinfo.iter_CUs():
            # DWARFInfo allows to iterate over the compile units contained in
            # the .debug_info section. CU is a CompileUnit object, with some
            # computed attributes (such as its offset in the section) and
            # a header which conforms to the DWARF standard. The access to
            # header elements is, as usual, via item-lookup.
            print('  Found a compile unit at offset %s, length %s' % (
                CU.cu_offset, CU['unit_length']))

            # A CU provides a simple API to iterate over all the DIEs in it.
            for DIE in CU.iter_DIEs():
                # Go over all attributes of the DIE. Each attribute is an
                # AttributeValue object (from elftools.dwarf.die), which we
                # can examine.
                for attr in DIE.attributes.values():
                    # Check if this attribute contains location information
                    if loc_parser.attribute_has_location(attr, CU['version']):
                        print('   DIE %s. attr %s.' % (DIE.tag, attr.name))
                        loc = loc_parser.parse_from_attribute(attr,
                                                              CU['version'], DIE)
                        # We either get a list (in case the attribute is a
                        # reference to the .debug_loc section) or a LocationExpr
                        # object (in case the attribute itself contains location
                        # information).
                        if isinstance(loc, LocationExpr):
                            print('      %s' % (
                                describe_DWARF_expr(loc.loc_expr,
                                                    dwarfinfo.structs, CU.cu_offset)))
                        elif isinstance(loc, list):
                            print(show_loclist(loc,
                                               dwarfinfo,
                                               '      ', CU.cu_offset))

def show_loclist(loclist, dwarfinfo, indent, cu_offset):
    """ Display a location list nicely, decoding the DWARF expressions
        contained within.
    """
    d = []
    for loc_entity in loclist:
        if isinstance(loc_entity, LocationEntry):
            d.append('%s <<%s>>' % (
                loc_entity,
                describe_DWARF_expr(loc_entity.loc_expr, dwarfinfo.structs, cu_offset)))
        else:
            d.append(str(loc_entity))
    return '\n'.join(indent + s for s in d)

if __name__ == '__main__':
    if sys.argv[1] == '--test':
        for filename in sys.argv[2:]:
            process_file(filename)



#-------------------------------------------------------------------------------
# elftools example: elfclass_address_size.py
#
# This example explores the ELF class (32 or 64-bit) and address size in each
# of the CUs in the DWARF information.
#
# Eli Bendersky (eliben@gmail.com)
# This code is in the public domain
#-------------------------------------------------------------------------------
from __future__ import print_function
import sys

# If pyelftools is not installed, the example can also run from the root or
# examples/ dir of the source distribution.
sys.path[0:0] = ['.', '..']

from elftools.elf.elffile import ELFFile


def process_file(filename):
    with open(filename, 'rb') as f:
        elffile = ELFFile(f)
        # elfclass is a public attribute of ELFFile, read from its header
        print('%s: elfclass is %s' % (filename, elffile.elfclass))

        if elffile.has_dwarf_info():
            dwarfinfo = elffile.get_dwarf_info()
            for CU in dwarfinfo.iter_CUs():
                # cu_offset is a public attribute of CU
                # address_size is part of the CU header
                print('  CU at offset 0x%x. address_size is %s' % (
                    CU.cu_offset, CU['address_size']))


if __name__ == '__main__':
    if sys.argv[1] == '--test':
        for filename in sys.argv[2:]:
            process_file(filename)



from .binary import (
    int_to_bin, bin_to_int, swap_bytes, encode_bin, decode_bin)
from .bitstream import BitStreamReader, BitStreamWriter
from .container import (Container, FlagsContainer, ListContainer,
    LazyContainer)
from .hex import HexString, hexdump




#-------------------------------------------------------------------------------
# elftools: common/construct_utils.py
#
# Some complementary construct utilities
#
# Eli Bendersky (eliben@gmail.com)
# This code is in the public domain
#-------------------------------------------------------------------------------
from struct import Struct
from ..construct import (
    Subconstruct, ConstructError, ArrayError, Adapter, Field, RepeatUntil,
    Rename, SizeofError, Construct, StaticField, FieldError
    )


class RepeatUntilExcluding(Subconstruct):
    """ A version of construct's RepeatUntil that doesn't include the last
        element (which casued the repeat to exit) in the return value.

        Only parsing is currently implemented.

        P.S. removed some code duplication
    """
    __slots__ = ["predicate"]
    def __init__(self, predicate, subcon):
        Subconstruct.__init__(self, subcon)
        self.predicate = predicate
        self._clear_flag(self.FLAG_COPY_CONTEXT)
        self._set_flag(self.FLAG_DYNAMIC)
    def _parse(self, stream, context):
        obj = []
        try:
            context_for_subcon = context
            if self.subcon.conflags & self.FLAG_COPY_CONTEXT:
                context_for_subcon = context.__copy__()

            while True:
                subobj = self.subcon._parse(stream, context_for_subcon)
                if self.predicate(subobj, context):
                    break
                obj.append(subobj)
        except ConstructError as ex:
            raise ArrayError("missing terminator", ex)
        return obj
    def _build(self, obj, stream, context):
        raise NotImplementedError('no building')
    def _sizeof(self, context):
        raise SizeofError("can't calculate size")

class ULEB128(Construct):
    """A construct based parser for ULEB128 encoding.

       Incompatible with Python 2 - assumes that the return of read()
       is an indexed collection of numbers.
    """
    def _parse(self, stream, context):
        value = 0
        shift = 0
        while True:
            data = stream.read(1)
            if len(data) != 1:
                raise FieldError("unexpected end of stream while parsing a ULEB128 encoded value")
            b = data[0]
            value |= (b & 0x7F) << shift
            shift += 7
            if b & 0x80 == 0:
                return value

class SLEB128(Construct):
    """A construct based parser for SLEB128 encoding.

       Incompatible with Python 2 - assumes that the return of read()
       is an indexed collection of numbers.
    """
    def _parse(self, stream, context):
        value = 0
        shift = 0
        while True:
            data = stream.read(1)
            if len(data) != 1:
                raise FieldError("unexpected end of stream while parsing a SLEB128 encoded value")
            b = data[0]
            value |= (b & 0x7F) << shift
            shift += 7
            if b & 0x80 == 0:
                return value | (~0 << shift) if b & 0x40 else value

class StreamOffset(Construct):
    """
    Captures the current stream offset

    Parameters:
    * name - the name of the value

    Example:
    StreamOffset("item_offset")
    """
    __slots__ = []
    def __init__(self, name):
        Construct.__init__(self, name)
        self._set_flag(self.FLAG_DYNAMIC)
    def _parse(self, stream, context):
        return stream.tell()
    def _build(self, obj, stream, context):
        context[self.name] = stream.tell()
    def _sizeof(self, context):
        return 0

_UBInt24_packer = Struct(">BH")
_ULInt24_packer = Struct("<HB")

class UBInt24(StaticField):
    """unsigned, big endian 24-bit integer"""
    def __init__(self, name):
        StaticField.__init__(self, name, 3)

    def _parse(self, stream, context):
        (h, l) = _UBInt24_packer.unpack(StaticField._parse(self, stream, context))
        return l | (h << 16)
    
    def _build(self, obj, stream, context):
        StaticField._build(self, _UBInt24_packer.pack(obj >> 16, obj & 0xFFFF), stream, context)

class ULInt24(StaticField):
    """unsigned, little endian 24-bit integer"""
    def __init__(self, name):
        StaticField.__init__(self, name, 3)

    def _parse(self, stream, context):
        (l, h) = _ULInt24_packer.unpack(StaticField._parse(self, stream, context))
        return l | (h << 16)
    
    def _build(self, obj, stream, context):
        StaticField._build(self, _ULInt24_packer.pack(obj & 0xFFFF, obj >> 16), stream, context)



#-------------------------------------------------------------------------------
# elftools: elf/descriptions.py
#
# Textual descriptions of the various enums and flags of ELF
#
# Eli Bendersky (eliben@gmail.com)
# This code is in the public domain
#-------------------------------------------------------------------------------
from .enums import (
    ENUM_D_TAG, ENUM_E_VERSION, ENUM_P_TYPE_BASE, ENUM_SH_TYPE_BASE,
    ENUM_RELOC_TYPE_i386, ENUM_RELOC_TYPE_x64,
    ENUM_RELOC_TYPE_ARM, ENUM_RELOC_TYPE_AARCH64, ENUM_RELOC_TYPE_PPC64,
    ENUM_RELOC_TYPE_MIPS, ENUM_ATTR_TAG_ARM, ENUM_ATTR_TAG_RISCV,
    ENUM_RELOC_TYPE_S390X, ENUM_RELOC_TYPE_LOONGARCH, ENUM_DT_FLAGS,
    ENUM_DT_FLAGS_1, ENUM_RELOC_TYPE_PPC)
from .constants import (
    P_FLAGS, RH_FLAGS, SH_FLAGS, SUNW_SYMINFO_FLAGS, VER_FLAGS)
from ..common.utils import bytes2hex


def describe_ei_class(x):
    return _DESCR_EI_CLASS.get(x, _unknown)


def describe_ei_data(x):
    return _DESCR_EI_DATA.get(x, _unknown)


def describe_ei_version(x):
    s = '%d' % ENUM_E_VERSION[x]
    if x == 'EV_CURRENT':
        s += ' (current)'
    return s


def describe_ei_osabi(x):
    return _DESCR_EI_OSABI.get(x, _unknown)


def describe_e_type(x, elffile=None):
    if elffile is not None and x == 'ET_DYN':
        # Detect whether this is a normal SO or a PIE executable
        dynamic = elffile.get_section_by_name('.dynamic')
        for t in dynamic.iter_tags('DT_FLAGS_1'):
            if t.entry.d_val & ENUM_DT_FLAGS_1['DF_1_PIE']:
                return 'DYN (Position-Independent Executable file)'
    return _DESCR_E_TYPE.get(x, _unknown)


def describe_e_machine(x):
    return _DESCR_E_MACHINE.get(x, _unknown)


def describe_e_version_numeric(x):
    return '0x%x' % ENUM_E_VERSION[x]


def describe_p_type(x):
    if x in _DESCR_P_TYPE:
        return _DESCR_P_TYPE.get(x)
    elif x >= ENUM_P_TYPE_BASE['PT_LOOS'] and x <= ENUM_P_TYPE_BASE['PT_HIOS']:
        return 'LOOS+%lx' % (x - ENUM_P_TYPE_BASE['PT_LOOS'])
    else:
        return _unknown


def describe_p_flags(x):
    s = ''
    for flag in (P_FLAGS.PF_R, P_FLAGS.PF_W, P_FLAGS.PF_X):
        s += _DESCR_P_FLAGS[flag] if (x & flag) else ' '
    return s


def describe_rh_flags(x):
    return ' '.join(
        _DESCR_RH_FLAGS[flag]
        for flag in (RH_FLAGS.RHF_NONE, RH_FLAGS.RHF_QUICKSTART,
                     RH_FLAGS.RHF_NOTPOT, RH_FLAGS.RHF_NO_LIBRARY_REPLACEMENT,
                     RH_FLAGS.RHF_NO_MOVE, RH_FLAGS.RHF_SGI_ONLY,
                     RH_FLAGS.RHF_GUARANTEE_INIT,
                     RH_FLAGS.RHF_DELTA_C_PLUS_PLUS,
                     RH_FLAGS.RHF_GUARANTEE_START_INIT, RH_FLAGS.RHF_PIXIE,
                     RH_FLAGS.RHF_DEFAULT_DELAY_LOAD,
                     RH_FLAGS.RHF_REQUICKSTART, RH_FLAGS.RHF_REQUICKSTARTED,
                     RH_FLAGS.RHF_CORD, RH_FLAGS.RHF_NO_UNRES_UNDEF,
                     RH_FLAGS.RHF_RLD_ORDER_SAFE)
        if x & flag)


def describe_sh_type(x):
    if x in _DESCR_SH_TYPE:
        return _DESCR_SH_TYPE.get(x)
    elif (x >= ENUM_SH_TYPE_BASE['SHT_LOOS'] and
          x < ENUM_SH_TYPE_BASE['SHT_GNU_versym']):
        return 'loos+0x%lx' % (x - ENUM_SH_TYPE_BASE['SHT_LOOS'])
    else:
        return _unknown


def describe_sh_flags(x):
    s = ''
    for flag in (
            SH_FLAGS.SHF_WRITE, SH_FLAGS.SHF_ALLOC, SH_FLAGS.SHF_EXECINSTR,
            SH_FLAGS.SHF_MERGE, SH_FLAGS.SHF_STRINGS, SH_FLAGS.SHF_INFO_LINK,
            SH_FLAGS.SHF_LINK_ORDER, SH_FLAGS.SHF_OS_NONCONFORMING,
            SH_FLAGS.SHF_GROUP, SH_FLAGS.SHF_TLS, SH_FLAGS.SHF_MASKOS,
            SH_FLAGS.SHF_EXCLUDE):
        s += _DESCR_SH_FLAGS[flag] if (x & flag) else ''
    if not x & SH_FLAGS.SHF_EXCLUDE:
        if x & SH_FLAGS.SHF_MASKPROC:
            s += 'p'
    return s


def describe_symbol_type(x):
    return _DESCR_ST_INFO_TYPE.get(x, _unknown)


def describe_symbol_bind(x):
    return _DESCR_ST_INFO_BIND.get(x, _unknown)


def describe_symbol_visibility(x):
    return _DESCR_ST_VISIBILITY.get(x, _unknown)


def describe_symbol_local(x):
    return '[<localentry>: ' + str(1 << x) + ']'


def describe_symbol_other(x):
    vis = describe_symbol_visibility(x['visibility'])
    if x['local'] > 1 and x['local'] < 7:
        return vis + ' ' + describe_symbol_local(x['local'])
    return vis


def describe_symbol_shndx(x):
    return _DESCR_ST_SHNDX.get(x, '%3s' % x)


def describe_reloc_type(x, elffile):
    arch = elffile.get_machine_arch()
    if arch == 'x86':
        return _DESCR_RELOC_TYPE_i386.get(x, _unknown)
    elif arch == 'x64':
        return _DESCR_RELOC_TYPE_x64.get(x, _unknown)
    elif arch == 'ARM':
        return _DESCR_RELOC_TYPE_ARM.get(x, _unknown)
    elif arch == 'AArch64':
        return _DESCR_RELOC_TYPE_AARCH64.get(x, _unknown)
    elif arch == '64-bit PowerPC':
        return _DESCR_RELOC_TYPE_PPC64.get(x, _unknown)
    elif arch == 'PowerPC':
        return _DESCR_RELOC_TYPE_PPC.get(x, _unknown)
    elif arch == 'IBM S/390':
        return _DESCR_RELOC_TYPE_S390X.get(x, _unknown)
    elif arch == 'MIPS':
        return _DESCR_RELOC_TYPE_MIPS.get(x, _unknown)
    elif arch == 'LoongArch':
        return _DESCR_RELOC_TYPE_LOONGARCH.get(x, _unknown)
    else:
        return 'unrecognized: %-7x' % (x & 0xFFFFFFFF)


def describe_dyn_tag(x):
    return _DESCR_D_TAG.get(x, _unknown)


def describe_dt_flags(x):
    return ' '.join(key[3:] for key, val in
        sorted(ENUM_DT_FLAGS.items(), key=lambda t: t[1]) if x & val)


def describe_dt_flags_1(x):
    return ' '.join(key[5:] for key, val in
        sorted(ENUM_DT_FLAGS_1.items(), key=lambda t: t[1]) if x & val)


def describe_syminfo_flags(x):
    return ''.join(_DESCR_SYMINFO_FLAGS[flag] for flag in (
        SUNW_SYMINFO_FLAGS.SYMINFO_FLG_CAP,
        SUNW_SYMINFO_FLAGS.SYMINFO_FLG_DIRECT,
        SUNW_SYMINFO_FLAGS.SYMINFO_FLG_FILTER,
        SUNW_SYMINFO_FLAGS.SYMINFO_FLG_AUXILIARY,
        SUNW_SYMINFO_FLAGS.SYMINFO_FLG_DIRECTBIND,
        SUNW_SYMINFO_FLAGS.SYMINFO_FLG_COPY,
        SUNW_SYMINFO_FLAGS.SYMINFO_FLG_LAZYLOAD,
        SUNW_SYMINFO_FLAGS.SYMINFO_FLG_NOEXTDIRECT,
        SUNW_SYMINFO_FLAGS.SYMINFO_FLG_INTERPOSE,
        SUNW_SYMINFO_FLAGS.SYMINFO_FLG_DEFERRED) if x & flag)


def describe_symbol_boundto(x):
    return _DESCR_SYMINFO_BOUNDTO.get(x, '%3s' % x)


def describe_ver_flags(x):
    return ' | '.join(_DESCR_VER_FLAGS[flag] for flag in (
        VER_FLAGS.VER_FLG_WEAK,
        VER_FLAGS.VER_FLG_BASE,
        VER_FLAGS.VER_FLG_INFO) if x & flag)


def describe_note(x, machine):
    n_desc = x['n_desc']
    desc = ''
    if x['n_type'] == 'NT_GNU_ABI_TAG':
        if x['n_name'] == 'GNU':
            desc = '\n    OS: %s, ABI: %d.%d.%d' % (
                _DESCR_NOTE_ABI_TAG_OS.get(n_desc['abi_os'], _unknown),
                n_desc['abi_major'], n_desc['abi_minor'], n_desc['abi_tiny'])
        else:
            desc = '\n   description data: %s ' % bytes2hex(x['n_descdata'])
    elif x['n_type'] == 'NT_GNU_BUILD_ID':
        desc = '\n    Build ID: %s' % (n_desc)
    elif x['n_type'] == 'NT_GNU_GOLD_VERSION':
        desc = '\n    Version: %s' % (n_desc)
    elif x['n_type'] == 'NT_GNU_PROPERTY_TYPE_0':
        if x['n_name'] == 'GNU':
            desc = '\n      Properties: ' + describe_note_gnu_properties(x['n_desc'], machine)
        else:
            desc = '\n   description data: %s ' % bytes2hex(x['n_descdata'])
    else:
        desc = '\n      description data: {}'.format(bytes2hex(n_desc))

    if x['n_type'] == 'NT_GNU_ABI_TAG' and x['n_name'] == 'Android':
        note_type = 'NT_VERSION'
        note_type_desc = 'version'
    else:
        note_type = (x['n_type'] if isinstance(x['n_type'], str)
                    else 'Unknown note type:')
        note_type_desc = ('0x%.8x' % x['n_type']
                        if isinstance(x['n_type'], int) else
                        _DESCR_NOTE_N_TYPE.get(x['n_type'], _unknown))
    return '%s (%s)%s' % (note_type, note_type_desc, desc)


def describe_attr_tag_arm(tag, val, extra):
    idx = ENUM_ATTR_TAG_ARM[tag] - 1
    d_entry = _DESCR_ATTR_VAL_ARM[idx]

    if d_entry is None:
        if tag == 'TAG_COMPATIBILITY':
            return (_DESCR_ATTR_TAG_ARM[tag]
                    + 'flag = %d, vendor = %s' % (val, extra))

        elif tag == 'TAG_ALSO_COMPATIBLE_WITH':
            if val.tag == 'TAG_CPU_ARCH':
                return _DESCR_ATTR_TAG_ARM[tag] + d_entry[val]

            else:
                return _DESCR_ATTR_TAG_ARM[tag] + '??? (%d)' % val.tag

        elif tag == 'TAG_NODEFAULTS':
            return _DESCR_ATTR_TAG_ARM[tag] + 'True'

        s = _DESCR_ATTR_TAG_ARM[tag]
        s += '"%s"' % val if val else ''
        return s

    else:
        return _DESCR_ATTR_TAG_ARM[tag] + d_entry[val]

def describe_attr_tag_riscv(tag, val, extra):
    idx = ENUM_ATTR_TAG_RISCV[tag] - 1
    d_entry = _DESCR_ATTR_VAL_RISCV[idx]

    if d_entry is None:
        s = _DESCR_ATTR_TAG_RISCV[tag]
        s += '"%s"' % val if val else ''
        return s

    else:
        return _DESCR_ATTR_TAG_RISCV[tag] + d_entry[val]

def describe_note_gnu_property_bitmap_and(values, prefix, value):
    descs = []
    for mask, desc in values:
        if value & mask:
            descs.append(desc)
    return '%s: %s' % (prefix, ', '.join(descs))

def describe_note_gnu_properties(properties, machine):
    descriptions = []
    for prop in properties:
        t, d, sz = prop.pr_type, prop.pr_data, prop.pr_datasz
        if t == 'GNU_PROPERTY_STACK_SIZE':
            if type(d) is int:
                prop_desc = 'stack size: 0x%x' % d
            else:
                prop_desc = 'stack size: <corrupt length: 0x%x>' % sz
        elif t == 'GNU_PROPERTY_NO_COPY_ON_PROTECTED':
            if sz != 0:
                prop_desc = ' <corrupt length: 0x%x>' % sz
            else:
                prop_desc = 'no copy on protected'
        elif t == 'GNU_PROPERTY_X86_FEATURE_1_AND':
            if sz != 4:
                prop_desc = ' <corrupt length: 0x%x>' % sz
            else:
                prop_desc = describe_note_gnu_property_bitmap_and(_DESCR_NOTE_GNU_PROPERTY_X86_FEATURE_1_FLAGS, 'x86 feature', d)
        elif t == 'GNU_PROPERTY_X86_FEATURE_2_USED':
            if sz != 4:
                prop_desc = ' <corrupt length: 0x%x>' % sz
            else:
                prop_desc = describe_note_gnu_property_bitmap_and(_DESCR_NOTE_GNU_PROPERTY_X86_FEATURE_2_FLAGS, 'x86 feature used', d)                
        elif t == 'GNU_PROPERTY_X86_ISA_1_NEEDED':
            if sz != 4:
                prop_desc = ' <corrupt length: 0x%x>' % sz
            else:
                prop_desc = describe_note_gnu_property_bitmap_and(_DESCR_NOTE_GNU_PROPERTY_X86_ISA_1_FLAGS, 'x86 ISA needed', d)
        elif t == 'GNU_PROPERTY_X86_ISA_1_USED':
            if sz != 4:
                prop_desc = ' <corrupt length: 0x%x>' % sz
            else:
                prop_desc = describe_note_gnu_property_bitmap_and(_DESCR_NOTE_GNU_PROPERTY_X86_ISA_1_FLAGS, 'x86 ISA used', d)
        elif t == 'GNU_PROPERTY_AARCH64_FEATURE_1_AND' and machine == 'EM_AARCH64':
            if sz != 4:
                prop_desc = ' <corrupt length: 0x%x>' % sz
            else:
                prop_desc = describe_note_gnu_property_bitmap_and(_DESCR_NOTE_GNU_PROPERTY_AARCH64_FEATURE_1_AND, 'aarch64 feature', d)
        elif t == 'GNU_PROPERTY_AARCH64_FEATURE_1_AND' and machine == 'EM_RISCV':
            # RISC-V shares the same bit-mask with AArch64
            if sz != 4:
                prop_desc = ' <corrupt length: 0x%x>' % sz
            else:
                prop_desc = describe_note_gnu_property_bitmap_and(_DESCR_NOTE_GNU_PROPERTY_RISCV_FEATURE_1_AND, 'RISC-V AND feature', d)
        elif _DESCR_NOTE_GNU_PROPERTY_TYPE_LOPROC <= t <= _DESCR_NOTE_GNU_PROPERTY_TYPE_HIPROC:
            prop_desc = '<processor-specific type 0x%x data: %s >' % (t, bytes2hex(d, sep=' '))
        elif _DESCR_NOTE_GNU_PROPERTY_TYPE_LOUSER <= t <= _DESCR_NOTE_GNU_PROPERTY_TYPE_HIUSER:
            prop_desc = '<application-specific type 0x%x data: %s >' % (t, bytes2hex(d, sep=' '))
        else:
            prop_desc = '<unknown type 0x%x data: %s >' % (t, bytes2hex(d, sep=' '))
        descriptions.append(prop_desc)
    return '\n        '.join(descriptions)

#-------------------------------------------------------------------------------
_unknown = '<unknown>'


_DESCR_EI_CLASS = dict(
    ELFCLASSNONE='none',
    ELFCLASS32='ELF32',
    ELFCLASS64='ELF64',
)


_DESCR_EI_DATA = dict(
    ELFDATANONE='none',
    ELFDATA2LSB="2's complement, little endian",
    ELFDATA2MSB="2's complement, big endian",
)


_DESCR_EI_OSABI = dict(
    ELFOSABI_SYSV='UNIX - System V',
    ELFOSABI_HPUX='UNIX - HP-UX',
    ELFOSABI_NETBSD='UNIX - NetBSD',
    ELFOSABI_LINUX='UNIX - Linux',
    ELFOSABI_HURD='UNIX - GNU/Hurd',
    ELFOSABI_SOLARIS='UNIX - Solaris',
    ELFOSABI_AIX='UNIX - AIX',
    ELFOSABI_IRIX='UNIX - IRIX',
    ELFOSABI_FREEBSD='UNIX - FreeBSD',
    ELFOSABI_TRU64='UNIX - TRU64',
    ELFOSABI_MODESTO='Novell - Modesto',
    ELFOSABI_OPENBSD='UNIX - OpenBSD',
    ELFOSABI_OPENVMS='VMS - OpenVMS',
    ELFOSABI_NSK='HP - Non-Stop Kernel',
    ELFOSABI_AROS='AROS',
    ELFOSABI_FENIXOS='Fenix OS',
    ELFOSABI_CLOUD='Nuxi - CloudABI',
    ELFOSABI_SORTIX='Sortix',
    ELFOSABI_ARM_AEABI='ARM - EABI',
    ELFOSABI_ARM='ARM - ABI',
    ELFOSABI_CELL_LV2='CellOS Lv-2',
    ELFOSABI_STANDALONE='Standalone App',
)


_DESCR_E_TYPE = dict(
    ET_NONE='NONE (None)',
    ET_REL='REL (Relocatable file)',
    ET_EXEC='EXEC (Executable file)',
    ET_DYN='DYN (Shared object file)',
    ET_CORE='CORE (Core file)',
    PROC_SPECIFIC='Processor Specific',
)


_DESCR_E_MACHINE = dict(
    EM_NONE='None',
    EM_M32='WE32100',
    EM_SPARC='Sparc',
    EM_386='Intel 80386',
    EM_68K='MC68000',
    EM_88K='MC88000',
    EM_860='Intel 80860',
    EM_MIPS='MIPS R3000',
    EM_S370='IBM System/370',
    EM_S390='IBM S/390',
    EM_MIPS_RS4_BE='MIPS 4000 big-endian',
    EM_IA_64='Intel IA-64',
    EM_X86_64='Advanced Micro Devices X86-64',
    EM_AVR='Atmel AVR 8-bit microcontroller',
    EM_ARM='ARM',
    EM_AARCH64='AArch64',
    EM_BLACKFIN='Analog Devices Blackfin',
    EM_PPC='PowerPC',
    EM_PPC64='PowerPC64',
    EM_RISCV='RISC-V',
    EM_LOONGARCH='LoongArch',
    RESERVED='RESERVED',
)


_DESCR_P_TYPE = dict(
    PT_NULL='NULL',
    PT_LOAD='LOAD',
    PT_DYNAMIC='DYNAMIC',
    PT_INTERP='INTERP',
    PT_NOTE='NOTE',
    PT_SHLIB='SHLIB',
    PT_PHDR='PHDR',
    PT_GNU_EH_FRAME='GNU_EH_FRAME',
    PT_GNU_STACK='GNU_STACK',
    PT_GNU_RELRO='GNU_RELRO',
    PT_GNU_PROPERTY='GNU_PROPERTY',
    PT_ARM_ARCHEXT='ARM_ARCHEXT',
    PT_ARM_EXIDX='EXIDX',  # binutils calls this EXIDX, not ARM_EXIDX
    PT_AARCH64_ARCHEXT='AARCH64_ARCHEXT',
    PT_AARCH64_UNWIND='AARCH64_UNWIND',
    PT_TLS='TLS',
    PT_MIPS_ABIFLAGS='ABIFLAGS',
    PT_RISCV_ATTRIBUTES='RISCV_ATTRIBUT',
)


_DESCR_P_FLAGS = {
    P_FLAGS.PF_X: 'E',
    P_FLAGS.PF_R: 'R',
    P_FLAGS.PF_W: 'W',
}


_DESCR_SH_TYPE = dict(
    SHT_NULL='NULL',
    SHT_PROGBITS='PROGBITS',
    SHT_SYMTAB='SYMTAB',
    SHT_STRTAB='STRTAB',
    SHT_RELA='RELA',
    SHT_HASH='HASH',
    SHT_DYNAMIC='DYNAMIC',
    SHT_NOTE='NOTE',
    SHT_NOBITS='NOBITS',
    SHT_REL='REL',
    SHT_SHLIB='SHLIB',
    SHT_DYNSYM='DYNSYM',
    SHT_INIT_ARRAY='INIT_ARRAY',
    SHT_FINI_ARRAY='FINI_ARRAY',
    SHT_PREINIT_ARRAY='PREINIT_ARRAY',
    SHT_GNU_ATTRIBUTES='GNU_ATTRIBUTES',
    SHT_GNU_HASH='GNU_HASH',
    SHT_GROUP='GROUP',
    SHT_SYMTAB_SHNDX='SYMTAB SECTION INDICIES',
    SHT_RELR='RELR',
    SHT_GNU_verdef='VERDEF',
    SHT_GNU_verneed='VERNEED',
    SHT_GNU_versym='VERSYM',
    SHT_GNU_LIBLIST='GNU_LIBLIST',
    SHT_AMD64_UNWIND='X86_64_UNWIND',
    SHT_ARM_EXIDX='ARM_EXIDX',
    SHT_ARM_PREEMPTMAP='ARM_PREEMPTMAP',
    SHT_ARM_ATTRIBUTES='ARM_ATTRIBUTES',
    SHT_ARM_DEBUGOVERLAY='ARM_DEBUGOVERLAY',
    SHT_AARCH64_ATTRIBUTES='AARCH64_ATTRIBUTES',
    SHT_RISCV_ATTRIBUTES='RISCV_ATTRIBUTES',
    SHT_MIPS_LIBLIST='MIPS_LIBLIST',
    SHT_MIPS_DEBUG='MIPS_DEBUG',
    SHT_MIPS_REGINFO='MIPS_REGINFO',
    SHT_MIPS_PACKAGE='MIPS_PACKAGE',
    SHT_MIPS_PACKSYM='MIPS_PACKSYM',
    SHT_MIPS_RELD='MIPS_RELD',
    SHT_MIPS_IFACE='MIPS_IFACE',
    SHT_MIPS_CONTENT='MIPS_CONTENT',
    SHT_MIPS_OPTIONS='MIPS_OPTIONS',
    SHT_MIPS_SHDR='MIPS_SHDR',
    SHT_MIPS_FDESC='MIPS_FDESC',
    SHT_MIPS_EXTSYM='MIPS_EXTSYM',
    SHT_MIPS_DENSE='MIPS_DENSE',
    SHT_MIPS_PDESC='MIPS_PDESC',
    SHT_MIPS_LOCSYM='MIPS_LOCSYM',
    SHT_MIPS_AUXSYM='MIPS_AUXSYM',
    SHT_MIPS_OPTSYM='MIPS_OPTSYM',
    SHT_MIPS_LOCSTR='MIPS_LOCSTR',
    SHT_MIPS_LINE='MIPS_LINE',
    SHT_MIPS_RFDESC='MIPS_RFDESC',
    SHT_MIPS_DELTASYM='MIPS_DELTASYM',
    SHT_MIPS_DELTAINST='MIPS_DELTAINST',
    SHT_MIPS_DELTACLASS='MIPS_DELTACLASS',
    SHT_MIPS_DWARF='MIPS_DWARF',
    SHT_MIPS_DELTADECL='MIPS_DELTADECL',
    SHT_MIPS_SYMBOL_LIB='MIPS_SYMBOL_LIB',
    SHT_MIPS_EVENTS='MIPS_EVENTS',
    SHT_MIPS_TRANSLATE='MIPS_TRANSLATE',
    SHT_MIPS_PIXIE='MIPS_PIXIE',
    SHT_MIPS_XLATE='MIPS_XLATE',
    SHT_MIPS_XLATE_DEBUG='MIPS_XLATE_DEBUG',
    SHT_MIPS_WHIRL='MIPS_WHIRL',
    SHT_MIPS_EH_REGION='MIPS_EH_REGION',
    SHT_MIPS_XLATE_OLD='MIPS_XLATE_OLD',
    SHT_MIPS_PDR_EXCEPTION='MIPS_PDR_EXCEPTION',
    SHT_MIPS_ABIFLAGS='MIPS_ABIFLAGS',
)


_DESCR_SH_FLAGS = {
    SH_FLAGS.SHF_WRITE: 'W',
    SH_FLAGS.SHF_ALLOC: 'A',
    SH_FLAGS.SHF_EXECINSTR: 'X',
    SH_FLAGS.SHF_MERGE: 'M',
    SH_FLAGS.SHF_STRINGS: 'S',
    SH_FLAGS.SHF_INFO_LINK: 'I',
    SH_FLAGS.SHF_LINK_ORDER: 'L',
    SH_FLAGS.SHF_OS_NONCONFORMING: 'O',
    SH_FLAGS.SHF_GROUP: 'G',
    SH_FLAGS.SHF_TLS: 'T',
    SH_FLAGS.SHF_MASKOS: 'o',
    SH_FLAGS.SHF_EXCLUDE: 'E',
}


_DESCR_RH_FLAGS = {
    RH_FLAGS.RHF_NONE: 'NONE',
    RH_FLAGS.RHF_QUICKSTART: 'QUICKSTART',
    RH_FLAGS.RHF_NOTPOT: 'NOTPOT',
    RH_FLAGS.RHF_NO_LIBRARY_REPLACEMENT: 'NO_LIBRARY_REPLACEMENT',
    RH_FLAGS.RHF_NO_MOVE: 'NO_MOVE',
    RH_FLAGS.RHF_SGI_ONLY: 'SGI_ONLY',
    RH_FLAGS.RHF_GUARANTEE_INIT: 'GUARANTEE_INIT',
    RH_FLAGS.RHF_DELTA_C_PLUS_PLUS: 'DELTA_C_PLUS_PLUS',
    RH_FLAGS.RHF_GUARANTEE_START_INIT: 'GUARANTEE_START_INIT',
    RH_FLAGS.RHF_PIXIE: 'PIXIE',
    RH_FLAGS.RHF_DEFAULT_DELAY_LOAD: 'DEFAULT_DELAY_LOAD',
    RH_FLAGS.RHF_REQUICKSTART: 'REQUICKSTART',
    RH_FLAGS.RHF_REQUICKSTARTED: 'REQUICKSTARTED',
    RH_FLAGS.RHF_CORD: 'CORD',
    RH_FLAGS.RHF_NO_UNRES_UNDEF: 'NO_UNRES_UNDEF',
    RH_FLAGS.RHF_RLD_ORDER_SAFE: 'RLD_ORDER_SAFE',
}


_DESCR_ST_INFO_TYPE = dict(
    STT_NOTYPE='NOTYPE',
    STT_OBJECT='OBJECT',
    STT_FUNC='FUNC',
    STT_SECTION='SECTION',
    STT_FILE='FILE',
    STT_COMMON='COMMON',
    STT_TLS='TLS',
    STT_NUM='NUM',
    STT_RELC='RELC',
    STT_SRELC='SRELC',
)


_DESCR_ST_INFO_BIND = dict(
    STB_LOCAL='LOCAL',
    STB_GLOBAL='GLOBAL',
    STB_WEAK='WEAK',
)


_DESCR_ST_VISIBILITY = dict(
    STV_DEFAULT='DEFAULT',
    STV_INTERNAL='INTERNAL',
    STV_HIDDEN='HIDDEN',
    STV_PROTECTED='PROTECTED',
    STV_EXPORTED='EXPORTED',
    STV_SINGLETON='SINGLETON',
    STV_ELIMINATE='ELIMINATE',
)


_DESCR_ST_SHNDX = dict(
    SHN_UNDEF='UND',
    SHN_ABS='ABS',
    SHN_COMMON='COM',
)


_DESCR_SYMINFO_FLAGS = {
    SUNW_SYMINFO_FLAGS.SYMINFO_FLG_DIRECT: 'D',
    SUNW_SYMINFO_FLAGS.SYMINFO_FLG_DIRECTBIND: 'B',
    SUNW_SYMINFO_FLAGS.SYMINFO_FLG_COPY: 'C',
    SUNW_SYMINFO_FLAGS.SYMINFO_FLG_LAZYLOAD: 'L',
    SUNW_SYMINFO_FLAGS.SYMINFO_FLG_NOEXTDIRECT: 'N',
    SUNW_SYMINFO_FLAGS.SYMINFO_FLG_AUXILIARY: 'A',
    SUNW_SYMINFO_FLAGS.SYMINFO_FLG_FILTER: 'F',
    SUNW_SYMINFO_FLAGS.SYMINFO_FLG_INTERPOSE: 'I',
    SUNW_SYMINFO_FLAGS.SYMINFO_FLG_CAP: 'S',
    SUNW_SYMINFO_FLAGS.SYMINFO_FLG_DEFERRED: 'P',
}


_DESCR_SYMINFO_BOUNDTO = dict(
    SYMINFO_BT_SELF='<self>',
    SYMINFO_BT_PARENT='<parent>',
    SYMINFO_BT_NONE='',
    SYMINFO_BT_EXTERN='<extern>',
)


_DESCR_VER_FLAGS = {
    0: '',
    VER_FLAGS.VER_FLG_BASE: 'BASE',
    VER_FLAGS.VER_FLG_WEAK: 'WEAK',
    VER_FLAGS.VER_FLG_INFO: 'INFO',
}


# PT_NOTE section types
_DESCR_NOTE_N_TYPE = dict(
    NT_GNU_ABI_TAG='ABI version tag',
    NT_GNU_HWCAP='DSO-supplied software HWCAP info',
    NT_GNU_BUILD_ID='unique build ID bitstring',
    NT_GNU_GOLD_VERSION='gold version',
    NT_GNU_PROPERTY_TYPE_0='program properties'
)


# Values in GNU .note.ABI-tag notes (n_type=='NT_GNU_ABI_TAG')
_DESCR_NOTE_ABI_TAG_OS = dict(
    ELF_NOTE_OS_LINUX='Linux',
    ELF_NOTE_OS_GNU='GNU',
    ELF_NOTE_OS_SOLARIS2='Solaris 2',
    ELF_NOTE_OS_FREEBSD='FreeBSD',
    ELF_NOTE_OS_NETBSD='NetBSD',
    ELF_NOTE_OS_SYLLABLE='Syllable',
)


# Values in GNU .note.gnu.property notes (n_type=='NT_GNU_PROPERTY_TYPE_0') have
# different formats which need to be parsed/described differently
_DESCR_NOTE_GNU_PROPERTY_TYPE_LOPROC=0xc0000000
_DESCR_NOTE_GNU_PROPERTY_TYPE_HIPROC=0xdfffffff
_DESCR_NOTE_GNU_PROPERTY_TYPE_LOUSER=0xe0000000
_DESCR_NOTE_GNU_PROPERTY_TYPE_HIUSER=0xffffffff


# Bit masks for GNU_PROPERTY_X86_FEATURE_1_xxx flags in the form
# (mask, flag_description) in the desired output order
_DESCR_NOTE_GNU_PROPERTY_X86_FEATURE_1_FLAGS = (
    (1, 'IBT'),
    (2, 'SHSTK'),
    (4, 'LAM_U48'),
    (8, 'LAM_U57'),
)

# Bit masks for GNU_PROPERTY_X86_FEATURE_2_xxx flags in the form
# (mask, flag_description) in the desired output order
_DESCR_NOTE_GNU_PROPERTY_X86_FEATURE_2_FLAGS = (
    (1, 'x86'),
    (2, 'x87'),
    (4, 'MMX'),
    (8, 'XMM'),
    (16, 'YMM'),
    (32, 'ZMM'),
)

# Same for GNU_PROPERTY_X86_SET_1_xxx
_DESCR_NOTE_GNU_PROPERTY_X86_ISA_1_FLAGS = (
    (1, 'x86-64-baseline'),
    # TODO; there is a long list
)

# Same for GNU_PROPERTY_AARCH64_FEATURE_1_AND
_DESCR_NOTE_GNU_PROPERTY_AARCH64_FEATURE_1_AND = (
    (1, 'bti'),
    (2, 'pac'),
)

_DESCR_NOTE_GNU_PROPERTY_RISCV_FEATURE_1_AND = (
    (1, 'ZICFILP'),
    (2, 'ZICFISS'),
)

def _reverse_dict(d, low_priority=()):
    """
    This is a tiny helper function to "reverse" the keys/values of a dictionary
    provided in the first argument, i.e. {k: v} becomes {v: k}.

    The second argument (optional) provides primitive control over what to do in
    the case of conflicting values - if a value is present in this list, it will
    not override any other entries of the same value.
    """
    out = {}
    for k, v in d.items():
        if v in out and k in low_priority:
            continue
        out[v] = k
    return out

_DESCR_RELOC_TYPE_i386 = _reverse_dict(ENUM_RELOC_TYPE_i386)
_DESCR_RELOC_TYPE_x64 = _reverse_dict(ENUM_RELOC_TYPE_x64)
_DESCR_RELOC_TYPE_ARM = _reverse_dict(ENUM_RELOC_TYPE_ARM)
_DESCR_RELOC_TYPE_AARCH64 = _reverse_dict(ENUM_RELOC_TYPE_AARCH64)
_DESCR_RELOC_TYPE_PPC64 = _reverse_dict(ENUM_RELOC_TYPE_PPC64)
_DESCR_RELOC_TYPE_PPC = _reverse_dict(ENUM_RELOC_TYPE_PPC)
_DESCR_RELOC_TYPE_S390X = _reverse_dict(ENUM_RELOC_TYPE_S390X)
_DESCR_RELOC_TYPE_MIPS = _reverse_dict(ENUM_RELOC_TYPE_MIPS)
_DESCR_RELOC_TYPE_LOONGARCH = _reverse_dict(ENUM_RELOC_TYPE_LOONGARCH)

_low_priority_D_TAG = (
    # these are 'meta-tags' marking semantics of numeric ranges of the enum
    # they should not override other tags with the same numbers
    # see https://docs.oracle.com/cd/E23824_01/html/819-0690/chapter6-42444.html
    'DT_LOOS',
    'DT_HIOS',
    'DT_LOPROC',
    'DT_HIPROC',
    'DT_ENCODING',
)
_DESCR_D_TAG = _reverse_dict(ENUM_D_TAG, low_priority=_low_priority_D_TAG)

_DESCR_ATTR_TAG_ARM = dict(
    TAG_FILE='File Attributes',
    TAG_SECTION='Section Attributes:',
    TAG_SYMBOL='Symbol Attributes:',
    TAG_CPU_RAW_NAME='Tag_CPU_raw_name: ',
    TAG_CPU_NAME='Tag_CPU_name: ',
    TAG_CPU_ARCH='Tag_CPU_arch: ',
    TAG_CPU_ARCH_PROFILE='Tag_CPU_arch_profile: ',
    TAG_ARM_ISA_USE='Tag_ARM_ISA_use: ',
    TAG_THUMB_ISA_USE='Tag_Thumb_ISA_use: ',
    TAG_FP_ARCH='Tag_FP_arch: ',
    TAG_WMMX_ARCH='Tag_WMMX_arch: ',
    TAG_ADVANCED_SIMD_ARCH='Tag_Advanced_SIMD_arch: ',
    TAG_PCS_CONFIG='Tag_PCS_config: ',
    TAG_ABI_PCS_R9_USE='Tag_ABI_PCS_R9_use: ',
    TAG_ABI_PCS_RW_DATA='Tag_ABI_PCS_RW_use: ',
    TAG_ABI_PCS_RO_DATA='Tag_ABI_PCS_RO_use: ',
    TAG_ABI_PCS_GOT_USE='Tag_ABI_PCS_GOT_use: ',
    TAG_ABI_PCS_WCHAR_T='Tag_ABI_PCS_wchar_t: ',
    TAG_ABI_FP_ROUNDING='Tag_ABI_FP_rounding: ',
    TAG_ABI_FP_DENORMAL='Tag_ABI_FP_denormal: ',
    TAG_ABI_FP_EXCEPTIONS='Tag_ABI_FP_exceptions: ',
    TAG_ABI_FP_USER_EXCEPTIONS='Tag_ABI_FP_user_exceptions: ',
    TAG_ABI_FP_NUMBER_MODEL='Tag_ABI_FP_number_model: ',
    TAG_ABI_ALIGN_NEEDED='Tag_ABI_align_needed: ',
    TAG_ABI_ALIGN_PRESERVED='Tag_ABI_align_preserved: ',
    TAG_ABI_ENUM_SIZE='Tag_ABI_enum_size: ',
    TAG_ABI_HARDFP_USE='Tag_ABI_HardFP_use: ',
    TAG_ABI_VFP_ARGS='Tag_ABI_VFP_args: ',
    TAG_ABI_WMMX_ARGS='Tag_ABI_WMMX_args: ',
    TAG_ABI_OPTIMIZATION_GOALS='Tag_ABI_optimization_goals: ',
    TAG_ABI_FP_OPTIMIZATION_GOALS='Tag_ABI_FP_optimization_goals: ',
    TAG_COMPATIBILITY='Tag_compatibility: ',
    TAG_CPU_UNALIGNED_ACCESS='Tag_CPU_unaligned_access: ',
    TAG_FP_HP_EXTENSION='Tag_FP_HP_extension: ',
    TAG_ABI_FP_16BIT_FORMAT='Tag_ABI_FP_16bit_format: ',
    TAG_MPEXTENSION_USE='Tag_MPextension_use: ',
    TAG_DIV_USE='Tag_DIV_use: ',
    TAG_NODEFAULTS='Tag_nodefaults: ',
    TAG_ALSO_COMPATIBLE_WITH='Tag_also_compatible_with: ',
    TAG_T2EE_USE='Tag_T2EE_use: ',
    TAG_CONFORMANCE='Tag_conformance: ',
    TAG_VIRTUALIZATION_USE='Tag_Virtualization_use: ',
    TAG_MPEXTENSION_USE_OLD='Tag_MPextension_use_old: ',
)

_DESCR_ATTR_VAL_ARM = [
    None, #1
    None, #2
    None, #3
    None, #4
    None, #5
    { #6 TAG_CPU_ARCH
        0 : 'Pre-v4',
        1 : 'v4',
        2 : 'v4T',
        3 : 'v5T',
        4 : 'v5TE',
        5 : 'v5TEJ',
        6 : 'v6',
        7 : 'v6KZ',
        8 : 'v6T2',
        9 : 'v6K',
        10: 'v7',
        11: 'v6-M',
        12: 'v6S-M',
        13: 'v7E-M',
        14: 'v8',
        15: 'v8-R',
        16: 'v8-M.baseline',
        17: 'v8-M.mainline',
    },
    { #7 TAG_CPU_ARCH_PROFILE
        0x00: 'None',
        0x41: 'Application',
        0x52: 'Realtime',
        0x4D: 'Microcontroller',
        0x53: 'Application or Realtime',
    },
    { #8 TAG_ARM_ISA
        0: 'No',
        1: 'Yes',
    },
    { #9 TAG_THUMB_ISA
        0: 'No',
        1: 'Thumb-1',
        2: 'Thumb-2',
        3: 'Yes',
    },
    { #10 TAG_FP_ARCH
        0: 'No',
        1: 'VFPv1',
        2: 'VFPv2 ',
        3: 'VFPv3',
        4: 'VFPv3-D16',
        5: 'VFPv4',
        6: 'VFPv4-D16',
        7: 'FP ARM v8',
        8: 'FPv5/FP-D16 for ARMv8',
    },
    { #11 TAG_WMMX_ARCH
        0: 'No',
        1: 'WMMXv1',
        2: 'WMMXv2',
    },
    { #12 TAG_ADVANCED_SIMD_ARCH
        0: 'No',
        1: 'NEONv1',
        2: 'NEONv1 with Fused-MAC',
        3: 'NEON for ARMv8',
        4: 'NEON for ARMv8.1',
    },
    { #13 TAG_PCS_CONFIG
        0: 'None',
        1: 'Bare platform',
        2: 'Linux application',
        3: 'Linux DSO',
        4: 'PalmOS 2004',
        5: 'PalmOS (reserved)',
        6: 'SymbianOS 2004',
        7: 'SymbianOS (reserved)',
    },
    { #14 TAG_ABI_PCS_R9_USE
        0: 'v6',
        1: 'SB',
        2: 'TLS',
        3: 'Unused',
    },
    { #15 TAG_ABI_PCS_RW_DATA
        0: 'Absolute',
        1: 'PC-relative',
        2: 'SB-relative',
        3: 'None',
    },
    { #16 TAG_ABI_PCS_RO_DATA
        0: 'Absolute',
        1: 'PC-relative',
        2: 'None',
    },
    { #17 TAG_ABI_PCS_GOT_USE
        0: 'None',
        1: 'direct',
        2: 'GOT-indirect',
    },
    { #18 TAG_ABI_PCS_WCHAR_T
        0: 'None',
        1: '??? 1',
        2: '2',
        3: '??? 3',
        4: '4',
    },
    { #19 TAG_ABI_FP_ROUNDING
        0: 'Unused',
        1: 'Needed',
    },
    { #20 TAG_ABI_FP_DENORMAL
        0: 'Unused',
        1: 'Needed',
        2: 'Sign only',
    },
    { #21 TAG_ABI_FP_EXCEPTIONS
        0: 'Unused',
        1: 'Needed',
    },
    { #22 TAG_ABI_FP_USER_EXCEPTIONS
        0: 'Unused',
        1: 'Needed',
    },
    { #23 TAG_ABI_FP_NUMBER_MODEL
        0: 'Unused',
        1: 'Finite',
        2: 'RTABI',
        3: 'IEEE 754',
    },
    { #24 TAG_ABI_ALIGN_NEEDED
        0: 'None',
        1: '8-byte',
        2: '4-byte',
        3: '??? 3',
    },
    { #25 TAG_ABI_ALIGN_PRESERVED
        0: 'None',
        1: '8-byte, except leaf SP',
        2: '8-byte',
        3: '??? 3',
    },
    { #26 TAG_ABI_ENUM_SIZE
        0: 'Unused',
        1: 'small',
        2: 'int',
        3: 'forced to int',
    },
    { #27 TAG_ABI_HARDFP_USE
        0: 'As Tag_FP_arch',
        1: 'SP only',
        2: 'Reserved',
        3: 'Deprecated',
    },
    { #28 TAG_ABI_VFP_ARGS
        0: 'AAPCS',
        1: 'VFP registers',
        2: 'custom',
        3: 'compatible',
    },
    { #29 TAG_ABI_WMMX_ARGS
        0: 'AAPCS',
        1: 'WMMX registers',
        2: 'custom',
    },
    { #30 TAG_ABI_OPTIMIZATION_GOALS
        0: 'None',
        1: 'Prefer Speed',
        2: 'Aggressive Speed',
        3: 'Prefer Size',
        4: 'Aggressive Size',
        5: 'Prefer Debug',
        6: 'Aggressive Debug',
    },
    { #31 TAG_ABI_FP_OPTIMIZATION_GOALS
        0: 'None',
        1: 'Prefer Speed',
        2: 'Aggressive Speed',
        3: 'Prefer Size',
        4: 'Aggressive Size',
        5: 'Prefer Accuracy',
        6: 'Aggressive Accuracy',
    },
    { #32 TAG_COMPATIBILITY
        0: 'No',
        1: 'Yes',
    },
    None, #33
    { #34 TAG_CPU_UNALIGNED_ACCESS
        0: 'None',
        1: 'v6',
    },
    None, #35
    { #36 TAG_FP_HP_EXTENSION
        0: 'Not Allowed',
        1: 'Allowed',
    },
    None, #37
    { #38 TAG_ABI_FP_16BIT_FORMAT
        0: 'None',
        1: 'IEEE 754',
        2: 'Alternative Format',
    },
    None, #39
    None, #40
    None, #41
    { #42 TAG_MPEXTENSION_USE
        0: 'Not Allowed',
        1: 'Allowed',
    },
    None, #43
    { #44 TAG_DIV_USE
        0: 'Allowed in Thumb-ISA, v7-R or v7-M',
        1: 'Not allowed',
        2: 'Allowed in v7-A with integer division extension',
    },
    None, #45
    None, #46
    None, #47
    None, #48
    None, #49
    None, #50
    None, #51
    None, #52
    None, #53
    None, #54
    None, #55
    None, #56
    None, #57
    None, #58
    None, #59
    None, #60
    None, #61
    None, #62
    None, #63
    None, #64
    None, #65
    { #66 TAG_FP_HP_EXTENSION
        0: 'Not Allowed',
        1: 'Allowed',
    },
    None, #67
    { #68 TAG_VIRTUALIZATION_USE
        0: 'Not Allowed',
        1: 'TrustZone',
        2: 'Virtualization Extensions',
        3: 'TrustZone and Virtualization Extensions',
    },
    None, #69
    { #70 TAG_MPEXTENSION_USE_OLD
        0: 'Not Allowed',
        1: 'Allowed',
    },
]

_DESCR_ATTR_TAG_RISCV = dict(
    TAG_FILE='File Attributes',
    TAG_SECTION='Section Attributes:',
    TAG_SYMBOL='Symbol Attributes:',
    TAG_STACK_ALIGN='Tag_RISCV_stack_align: ',
    TAG_ARCH='Tag_RISCV_arch: ',
    TAG_UNALIGNED='Tag_RISCV_unaligned_access: ',
    TAG_PRIV_SPEC='Tag_RISCV_priv_spec: ',
    TAG_PRIV_SPEC_MINOR='Tag_RISCV_priv_spec_minor: ',
    TAG_PRIV_SPEC_REVISION='Tag_RISCV_priv_spec_revision: ',
    TAG_ATOMIC_ABI='Tag_RISCV_atomic_abi: ',
    TAG_X3_REG_USAGE='Tag_RISCV_x3_reg_usage: ',
)

_DESCR_ATTR_VAL_RISCV = [
    None, #1
    None, #2
    None, #3
    { #4 TAG_RISCV_stack_align
        4: '4-bytes',
        16: '16-bytes',
    },
    None, #5 TAG_RISCV_arch
    { #6 TAG_RISCV_unaligned_access
        0: 'Not Allowed',
        1: 'Allowed',
    },
    None, #8 TAG_RISCV_priv_spec
    None, #10 TAG_RISCV_priv_spec_minor
    None, #12 TAG_RISCV_priv_spec_revision
    { #14 TAG_RISCV_atomic_abi
        0: 'UNKNOWN: This object uses unknown atomic ABI.',
        1: 'A6C: This object uses the A6 classical atomic ABI.',
        2: 'A6S: This object uses the strengthened A6 ABI.',
        3: 'A7: This object uses the A7 atomic ABI.'
    },
    { #16 TAG_RISCV_x3_reg_usage
        0: 'This object uses x3 as a fixed register with unknown purpose.',
        1: 'This object uses x3 as the global pointer, for relaxation purposes.',
        2: 'This object uses x3 as the shadow stack pointer.',
        3: 'This object uses X3 as a temporary register.',
    },
]



#------------------------------------------------------------------------------
# elftools: elf/gnuversions.py
#
# ELF sections
#
# Yann Rouillard (yann@pleiades.fr.eu.org)
# This code is in the public domain
#------------------------------------------------------------------------------
from ..construct import CString
from ..common.utils import struct_parse, elf_assert
from .sections import Section, Symbol


class Version(object):
    """ Version object - representing a version definition or dependency
        entry from a "Version Needed" or a "Version Dependency" table section.

        This kind of entry contains a pointer to an array of auxiliary entries
        that store the information about version names or dependencies.
        These entries are not stored in this object and should be accessed
        through the appropriate method of a section object which will return
        an iterator of VersionAuxiliary objects.

        Similarly to Section objects, allows dictionary-like access to
        verdef/verneed entry
    """
    def __init__(self, entry, name=None):
        self.entry = entry
        self.name = name

    def __getitem__(self, name):
        """ Implement dict-like access to entry
        """
        return self.entry[name]


class VersionAuxiliary(object):
    """ Version Auxiliary object - representing an auxiliary entry of a version
        definition or dependency entry

        Similarly to Section objects, allows dictionary-like access to the
        verdaux/vernaux entry
    """
    def __init__(self, entry, name):
        self.entry = entry
        self.name = name

    def __getitem__(self, name):
        """ Implement dict-like access to entries
        """
        return self.entry[name]


class GNUVersionSection(Section):
    """ Common ancestor class for ELF SUNW|GNU Version Needed/Dependency
        sections class which contains shareable code
    """

    def __init__(self, header, name, elffile, stringtable,
                 field_prefix, version_struct, version_auxiliaries_struct):
        super(GNUVersionSection, self).__init__(header, name, elffile)
        self.stringtable = stringtable
        self.field_prefix = field_prefix
        self.version_struct = version_struct
        self.version_auxiliaries_struct = version_auxiliaries_struct

    def num_versions(self):
        """ Number of version entries in the section
        """
        return self['sh_info']

    def _field_name(self, name, auxiliary=False):
        """ Return the real field's name of version or a version auxiliary
            entry
        """
        middle = 'a_' if auxiliary else '_'
        return self.field_prefix + middle + name

    def _iter_version_auxiliaries(self, entry_offset, count):
        """ Yield all auxiliary entries of a version entry
        """
        name_field = self._field_name('name', auxiliary=True)
        next_field = self._field_name('next', auxiliary=True)

        for _ in range(count):
            entry = struct_parse(
                        self.version_auxiliaries_struct,
                        self.stream,
                        stream_pos=entry_offset)

            name = self.stringtable.get_string(entry[name_field])
            version_aux = VersionAuxiliary(entry, name)
            yield version_aux

            entry_offset += entry[next_field]

    def iter_versions(self):
        """ Yield all the version entries in the section
            Each time it returns the main version structure
            and an iterator to walk through its auxiliaries entries
        """
        aux_field = self._field_name('aux')
        count_field = self._field_name('cnt')
        next_field = self._field_name('next')

        entry_offset = self['sh_offset']
        for _ in range(self.num_versions()):
            entry = struct_parse(
                self.version_struct,
                self.stream,
                stream_pos=entry_offset)

            elf_assert(entry[count_field] > 0,
                'Expected number of version auxiliary entries (%s) to be > 0'
                'for the following version entry: %s' % (
                    count_field, str(entry)))

            version = Version(entry)
            aux_entries_offset = entry_offset + entry[aux_field]
            version_auxiliaries_iter = self._iter_version_auxiliaries(
                    aux_entries_offset, entry[count_field])

            yield version, version_auxiliaries_iter

            entry_offset += entry[next_field]


class GNUVerNeedSection(GNUVersionSection):
    """ ELF SUNW or GNU Version Needed table section.
        Has an associated StringTableSection that's passed in the constructor.
    """
    def __init__(self, header, name, elffile, stringtable):
        super(GNUVerNeedSection, self).__init__(
                header, name, elffile, stringtable, 'vn',
                elffile.structs.Elf_Verneed, elffile.structs.Elf_Vernaux)
        self._has_indexes = None

    def has_indexes(self):
        """ Return True if at least one version definition entry has an index
            that is stored in the vna_other field.
            This information is used for symbol versioning
        """
        if self._has_indexes is None:
            self._has_indexes = False
            for _, vernaux_iter in self.iter_versions():
                for vernaux in vernaux_iter:
                    if vernaux['vna_other']:
                        self._has_indexes = True
                        break

        return self._has_indexes

    def iter_versions(self):
        for verneed, vernaux in super(GNUVerNeedSection, self).iter_versions():
            verneed.name = self.stringtable.get_string(verneed['vn_file'])
            yield verneed, vernaux

    def get_version(self, index):
        """ Get the version information located at index #n in the table
            Return boths the verneed structure and the vernaux structure
            that contains the name of the version
        """
        for verneed, vernaux_iter in self.iter_versions():
            for vernaux in vernaux_iter:
                if vernaux['vna_other'] == index:
                    return verneed, vernaux

        return None


class GNUVerDefSection(GNUVersionSection):
    """ ELF SUNW or GNU Version Definition table section.
        Has an associated StringTableSection that's passed in the constructor.
    """
    def __init__(self, header, name, elffile, stringtable):
        super(GNUVerDefSection, self).__init__(
                header, name, elffile, stringtable, 'vd',
                elffile.structs.Elf_Verdef, elffile.structs.Elf_Verdaux)

    def get_version(self, index):
        """ Get the version information located at index #n in the table
            Return boths the verdef structure and an iterator to retrieve
            both the version names and dependencies in the form of
            verdaux entries
        """
        for verdef, verdaux_iter in self.iter_versions():
            if verdef['vd_ndx'] == index:
                return verdef, verdaux_iter

        return None


class GNUVerSymSection(Section):
    """ ELF SUNW or GNU Versym table section.
        Has an associated SymbolTableSection that's passed in the constructor.
    """
    def __init__(self, header, name, elffile, symboltable):
        super(GNUVerSymSection, self).__init__(header, name, elffile)
        self.symboltable = symboltable

    def num_symbols(self):
        """ Number of symbols in the table
        """
        return self['sh_size'] // self['sh_entsize']

    def get_symbol(self, n):
        """ Get the symbol at index #n from the table (Symbol object)
            It begins at 1 and not 0 since the first entry is used to
            store the current version of the syminfo table
        """
        # Grab the symbol's entry from the stream
        entry_offset = self['sh_offset'] + n * self['sh_entsize']
        entry = struct_parse(
            self.structs.Elf_Versym,
            self.stream,
            stream_pos=entry_offset)
        # Find the symbol name in the associated symbol table
        name = self.symboltable.get_symbol(n).name
        return Symbol(entry, name)

    def iter_symbols(self):
        """ Yield all the symbols in the table
        """
        for i in range(self.num_symbols()):
            yield self.get_symbol(i)
